/*******************************************************************************
  MMFAKEALSA: A library for ALSA emulation with Open Sound System (OSS) drivers

  Last revision: 5/8/2007

  (c) 2007 by Michael Minn 
  See www.michaelminn.com for contact info

  This library contains code derived from the 
  Advanced Linux Sound Architecture (ALSA) Driver

  (c) 1994-2003 by Jaroslav Kysela <perex@suse.cz>,
                   Abramo Bagnara <abramo@alsa-project.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of version 2 of the GNU General Public License as 
  published by the Free Software Foundation.

*******************************************************************************/

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

typedef enum _snd_pcm_stream {
	SND_PCM_STREAM_PLAYBACK = 0,	/** Playback stream */
	SND_PCM_STREAM_CAPTURE,		/** Capture stream */
	SND_PCM_STREAM_LAST = SND_PCM_STREAM_CAPTURE
} snd_pcm_stream_t;

typedef enum _snd_pcm_format {
	SND_PCM_FORMAT_UNKNOWN = -1,	/** Unknown */
	SND_PCM_FORMAT_S8 = 0,		/** Signed 8 bit */
	SND_PCM_FORMAT_U8,		/** Unsigned 8 bit */
	SND_PCM_FORMAT_S16_LE,		/** Signed 16 bit Little Endian */
	SND_PCM_FORMAT_S16_BE,		/** Signed 16 bit Big Endian */
	SND_PCM_FORMAT_U16_LE,		/** Unsigned 16 bit Little Endian */
	SND_PCM_FORMAT_U16_BE,		/** Unsigned 16 bit Big Endian */
	SND_PCM_FORMAT_S24_LE,		/** Signed 24 bit Little Endian */
	SND_PCM_FORMAT_S24_BE,		/** Signed 24 bit Big Endian */
	SND_PCM_FORMAT_U24_LE,		/** Unsigned 24 bit Little Endian */
	SND_PCM_FORMAT_U24_BE,		/** Unsigned 24 bit Big Endian */
	SND_PCM_FORMAT_S32_LE,		/** Signed 32 bit Little Endian */
	SND_PCM_FORMAT_S32_BE,		/** Signed 32 bit Big Endian */
	SND_PCM_FORMAT_U32_LE,		/** Unsigned 32 bit Little Endian */
	SND_PCM_FORMAT_U32_BE,		/** Unsigned 32 bit Big Endian */
	SND_PCM_FORMAT_FLOAT_LE,	/** Float 32 bit Little Endian, Range -1.0 to 1.0 */
	SND_PCM_FORMAT_FLOAT_BE,	/** Float 32 bit Big Endian, Range -1.0 to 1.0 */
	SND_PCM_FORMAT_FLOAT64_LE,	/** Float 64 bit Little Endian, Range -1.0 to 1.0 */
	SND_PCM_FORMAT_FLOAT64_BE,	/** Float 64 bit Big Endian, Range -1.0 to 1.0 */
	SND_PCM_FORMAT_IEC958_SUBFRAME_LE,	/** IEC-958 Little Endian */
	SND_PCM_FORMAT_IEC958_SUBFRAME_BE,	/** IEC-958 Big Endian */
	SND_PCM_FORMAT_MU_LAW,		/** Mu-Law */
	SND_PCM_FORMAT_A_LAW,		/** A-Law */
	SND_PCM_FORMAT_IMA_ADPCM,	/** Ima-ADPCM */
	SND_PCM_FORMAT_MPEG,		/** MPEG */
	SND_PCM_FORMAT_GSM,		/** GSM */
	SND_PCM_FORMAT_SPECIAL = 31,	/** Special */
	SND_PCM_FORMAT_S24_3LE = 32,	/** Signed 24bit Little Endian in 3bytes format */
	SND_PCM_FORMAT_S24_3BE,		/** Signed 24bit Big Endian in 3bytes format */
	SND_PCM_FORMAT_U24_3LE,		/** Unsigned 24bit Little Endian in 3bytes format */
	SND_PCM_FORMAT_U24_3BE,		/** Unsigned 24bit Big Endian in 3bytes format */
	SND_PCM_FORMAT_S20_3LE,		/** Signed 20bit Little Endian in 3bytes format */
	SND_PCM_FORMAT_S20_3BE,		/** Signed 20bit Big Endian in 3bytes format */
	SND_PCM_FORMAT_U20_3LE,		/** Unsigned 20bit Little Endian in 3bytes format */
	SND_PCM_FORMAT_U20_3BE,		/** Unsigned 20bit Big Endian in 3bytes format */
	SND_PCM_FORMAT_S18_3LE,		/** Signed 18bit Little Endian in 3bytes format */
	SND_PCM_FORMAT_S18_3BE,		/** Signed 18bit Big Endian in 3bytes format */
	SND_PCM_FORMAT_U18_3LE,		/** Unsigned 18bit Little Endian in 3bytes format */
	SND_PCM_FORMAT_U18_3BE,		/** Unsigned 18bit Big Endian in 3bytes format */
	SND_PCM_FORMAT_LAST = SND_PCM_FORMAT_U18_3BE,
} snd_pcm_format_t;

typedef enum _snd_pcm_access {
	SND_PCM_ACCESS_MMAP_INTERLEAVED = 0,	/** mmap access with simple interleaved channels */
	SND_PCM_ACCESS_MMAP_NONINTERLEAVED,	/** mmap access with simple non interleaved channels */
	SND_PCM_ACCESS_MMAP_COMPLEX,		/** mmap access with complex placement */
	SND_PCM_ACCESS_RW_INTERLEAVED,		/** snd_pcm_readi/snd_pcm_writei access */
	SND_PCM_ACCESS_RW_NONINTERLEAVED,	/** snd_pcm_readn/snd_pcm_writen access */
	SND_PCM_ACCESS_LAST = SND_PCM_ACCESS_RW_NONINTERLEAVED
} snd_pcm_access_t;

typedef enum _snd_pcm_state {
	SND_PCM_STATE_OPEN = 0,		/** Open */
	SND_PCM_STATE_SETUP,		/** Setup installed */ 
	SND_PCM_STATE_PREPARED,		/** Ready to start */
	SND_PCM_STATE_RUNNING,		/** Running */
	SND_PCM_STATE_XRUN,		/** Stopped: underrun (playback) or overrun (capture) detected */
	SND_PCM_STATE_DRAINING,		/** Draining: running (playback) or stopped (capture) */
	SND_PCM_STATE_PAUSED,		/** Paused */
	SND_PCM_STATE_SUSPENDED,	/** Hardware is suspended */
	SND_PCM_STATE_DISCONNECTED,	/** Hardware is disconnected */
	SND_PCM_STATE_LAST = SND_PCM_STATE_DISCONNECTED
} snd_pcm_state_t;

/** Unsigned frames quantity */
typedef unsigned long snd_pcm_uframes_t;

/** Signed frames quantity */
typedef long snd_pcm_sframes_t;

/** These are values for the snd_pcm_open() mode parameter */
#define SND_PCM_NONBLOCK		0x0001
#define SND_PCM_ASYNC			0x0002

#define MMFAKEALSA_SNDLIB_VERSION	"1.0.13"


/***************************************************
Debugging
***************************************************/ 

static FILE *error_file = NULL;

void mmfakealsa_dbg(char *format, ...)
{
	return;

	va_list ap;
	char message[512] = "";

	if (!format)
		return;

	va_start(ap, format);
	vsprintf(message, format, ap);
	va_end(ap);

	fprintf(stderr, "%s", message);

	/**
	if (!error_file)
		error_file = fopen("/home/michael/mmfakealsa.err", "w+");

	if (error_file)
		fwrite(message, strlen(message), 1, error_file);
	**/
}



/***************************************************
Versioning info
***************************************************/ 

const char *snd_asoundlib_version(void)
{
	mmfakealsa_dbg("snd_asoundlib_version() = %s\n", MMFAKEALSA_SNDLIB_VERSION);

	return MMFAKEALSA_SNDLIB_VERSION;
}


/***************************************************
snd_config_t is a configuration structure of some kind.
snd_config must be global (accessed by inline functions?)
See src/conf.c
***************************************************/ 

typedef struct _snd_config snd_config_t;
snd_config_t *snd_config = NULL;

int snd_config_update(void)
{
	return 0;
}

int snd_config_copy(snd_config_t **dst, snd_config_t *src)
{
	mmfakealsa_dbg("snd_config_copy()\n");
	return 0;
}

int snd_config_delete(snd_config_t *config)
{
	mmfakealsa_dbg("snd_config_delete()\n");
	return 0;
}

int snd_config_search(snd_config_t *config, const char *key, snd_config_t **result)
{
	mmfakealsa_dbg("snd_config_search(%s)\n", key);
	return -EINVAL;
}

int snd_config_get_ascii(const snd_config_t *config, char **ascii)
{
	mmfakealsa_dbg("snd_config_get_ascii()\n");
	return -EINVAL;
}

int snd_config_imake_integer(snd_config_t **config, const char *id, const long value)
{
	mmfakealsa_dbg("snd_config_imake_integer(%s, %d)\n", id, value);
	return -EINVAL;
}

int snd_config_substitute(snd_config_t *dst, snd_config_t *src)
{
	return -EINVAL;
}

int snd_config_add(snd_config_t *father, snd_config_t *leaf)
{
	return -EINVAL;
}

int snd_config_search_alias_hooks(snd_config_t *config, const char *base, const char *key, snd_config_t **result)
{
	return -EINVAL;
}


/***************************************************
ALSA error and message handling - not implemented for now
***************************************************/ 

typedef struct _snd_output snd_output_t;

struct _snd_output
{
	FILE *output;
};

int snd_output_stdio_attach(snd_output_t **outputp, FILE *fp, int _close)
{
	return 0;
}

int snd_output_close(snd_output_t *output)
{
	return 0;
}


typedef void (*snd_lib_error_handler_t)(const char *file, int line, const char *function, int err, const char *fmt, ...);

int snd_lib_error_set_handler(snd_lib_error_handler_t handler)
{
	mmfakealsa_dbg("snd_lib_error_set_handler()\n");

	return 0;
}


/***************************************************
Centralized sound card control.
Not needed by mplayer, xmms or Flash, so not implemented.
***************************************************/ 

typedef struct _snd_ctl snd_ctl_t;
typedef struct _snd_ctl_card_info snd_ctl_card_info_t;
typedef struct _snd_pcm_info snd_pcm_info_t;

struct _snd_ctl
{
	int dummy;
};

struct _snd_pcm_info
{
	int dummy;
};

int snd_ctl_open(snd_ctl_t **ctl, const char *name, int mode)
{

	mmfakealsa_dbg("snd_ctl_open(%s)\n", name);
	if (!ctl)	
		return -EINVAL;

	*ctl = calloc(sizeof(snd_ctl_t), 1);
	return 0;
}

int snd_ctl_close(snd_ctl_t *ctl)
{
	mmfakealsa_dbg("snd_ctl_close()\n");

	if (ctl)
		free(ctl);

	return 0;
}

int snd_ctl_card_info(snd_ctl_t *ctl, snd_ctl_card_info_t *info)
{
	mmfakealsa_dbg("snd_ctl_card_info()\n");
	return 0;
}

size_t snd_ctl_card_info_sizeof()
{
	mmfakealsa_dbg("snd_ctl_card_info_sizeof()\n");
        return 0;
}

const char *snd_ctl_card_info_get_driver(const snd_ctl_card_info_t *obj)
{
	return "default";	
}

const char *snd_ctl_card_info_get_name(const snd_ctl_card_info_t *obj)
{
	return "default";	
}

int snd_ctl_pcm_info(snd_ctl_t *ctl, snd_pcm_info_t *info)
{
	mmfakealsa_dbg("snd_ctl_pcm_info()\n");
	return 0;
}

int snd_ctl_pcm_next_device(snd_ctl_t *ctl, int *device)
{
	mmfakealsa_dbg("snd_ctl_pcm_next_device()\n");

	if (!device)
		return -EINVAL;

	// Assume only one device on each card
	if (*device < 0)
		*device = 0;
	else
		*device = -1;

	return 0;
}

typedef struct _snd_ctl_elem_id snd_ctl_elem_id_t;
typedef struct _snd_ctl_elem_value snd_ctl_elem_value_t;

typedef enum _snd_ctl_elem_iface 
{
        SND_CTL_ELEM_IFACE_CARD = 0,
        SND_CTL_ELEM_IFACE_HWDEP,
        SND_CTL_ELEM_IFACE_MIXER, 
        SND_CTL_ELEM_IFACE_PCM,
        SND_CTL_ELEM_IFACE_RAWMIDI,
        SND_CTL_ELEM_IFACE_TIMER,
        SND_CTL_ELEM_IFACE_SEQUENCER,
        SND_CTL_ELEM_IFACE_LAST = SND_CTL_ELEM_IFACE_SEQUENCER
} snd_ctl_elem_iface_t;


int snd_ctl_elem_read(snd_ctl_t *ctl, snd_ctl_elem_value_t *value)
{
	mmfakealsa_dbg("snd_ctl_elem_read()\n");
	return -EINVAL;
}

int snd_ctl_elem_write(snd_ctl_t *ctl, snd_ctl_elem_value_t *value)
{
	mmfakealsa_dbg("snd_ctl_elem_read()\n");
	return -EINVAL;
}


size_t snd_ctl_elem_id_sizeof(void)
{
	return 64; // arbitrary value
}

void snd_ctl_elem_id_set_device(snd_ctl_elem_id_t *obj, unsigned int val)
{
}

void snd_ctl_elem_id_set_index(snd_ctl_elem_id_t *obj, unsigned int val)
{
}

void snd_ctl_elem_id_set_interface(snd_ctl_elem_id_t *obj, snd_ctl_elem_iface_t val)
{
}

void snd_ctl_elem_id_set_name(snd_ctl_elem_id_t *obj, const char *val)
{
}

void snd_ctl_elem_id_set_numid(snd_ctl_elem_id_t *obj, unsigned int val)
{
}

void snd_ctl_elem_id_set_subdevice(snd_ctl_elem_id_t *obj, unsigned int val)
{
}

size_t snd_ctl_elem_value_sizeof(void)
{
	return 64; // arbitrary value
}

void snd_ctl_elem_value_set_id(snd_ctl_elem_value_t *obj, const snd_ctl_elem_id_t *ptr)
{
}

void snd_ctl_elem_value_set_numid(snd_ctl_elem_value_t *obj, unsigned int val)
{
}

void snd_ctl_elem_value_set_interface(snd_ctl_elem_value_t *obj, snd_ctl_elem_iface_t val)
{
}

void snd_ctl_elem_value_set_index(snd_ctl_elem_value_t *obj, unsigned int val)
{
}

void snd_ctl_elem_value_set_integer(snd_ctl_elem_value_t *obj, unsigned int idx, long val)
{
}

void snd_ctl_elem_value_set_name(snd_ctl_elem_value_t *obj, const char *val)
{
}

void snd_ctl_elem_value_set_enumerated(snd_ctl_elem_value_t *obj, unsigned int idx, unsigned int val)
{
}

const void * snd_ctl_elem_value_get_bytes(const snd_ctl_elem_value_t *obj)
{
	return NULL;
}

/*************************************************** 
High Level Control Interface
(not a clue)

typedef struct _snd_hctl_elem snd_hctl_elem_t;

int snd_hctl_elem_write(snd_hctl_elem_t *elem, snd_ctl_elem_value_t * value)
{
	return -EINVAL;
}
***************************************************/



      
/*************************************************** 
Management for multiple sound cards.
Not needed by mplayer, xmms or Flash, so not implemented.
***************************************************/

int snd_card_next(int *card)
{
	if (!card)
		return -EINVAL;

	if (*card < 0)
		*card = 0;
	else
		*card = -1;

	return 0;
}

int snd_card_load(int card)
{
	// Return 1 if driver is present, 0 otherwise. Only one card supported for now.
	if (card == 0)
		return 1; 
	else
		return 0;
}

int snd_card_get_name(int card, char **name)
{
	if (!name || (card != 0))
		return -EINVAL;

	*name = "default";
	return 0;
}

/***************************************************
Not sure what this is
***************************************************/

typedef struct _snd_hwdep snd_hwdep_t;

int snd_hwdep_open(snd_hwdep_t **hwdep, const char *name, int mode)
{
	mmfakealsa_dbg("snd_hwdep_open()\n");
	return -EINVAL;
}

int snd_hwdep_close(snd_hwdep_t *hwdep)
{
	mmfakealsa_dbg("snd_hwdep_close()\n");
	return -EINVAL;
}

int snd_hwdep_poll_descriptors(snd_hwdep_t *hwdep, struct pollfd *pfds, unsigned int space)
{
	mmfakealsa_dbg("snd_hwdep_poll_descriptors()\n");
	return -EINVAL;
}



/***************************************************
snd_pcm is the device handle. The contents are private 
to the library functions, but the memory space is apparently
pulled off the stack using alloca() and I had problems
with memory corruption segfaults whenever I tried to
actually do anything with the allocated structure.
Therefore this pointer is passed around but unused
and static variables are used for keeping track
of device parameters and status. There is also an
issue because snd_pcm_open() can be called multiple
times for a single device to open for both input 
and output (?)
***************************************************/ 

typedef struct _snd_pcm snd_pcm_t;

struct _snd_pcm
{
	int dummy;
};

static int mmfakealsa_descriptor = -1;
static int mmfakealsa_channels = 2;
static int mmfakealsa_buffer_frames = 4410;	// based on 60 URBs in mmusbaudio
static int mmfakealsa_period_frames = 0;	// period = frames between interrupts
static int mmfakealsa_pcm_state = -1;
static int mmfakealsa_oss_format = AFMT_S16_LE;
static int mmfakealsa_alsa_format = SND_PCM_FORMAT_S16_LE;
static int mmfakealsa_sample_rate = 44100;
static int mmfakealsa_bits_per_sample = 16;
static char *mmfakealsa_oss_device = "/dev/dsp";

// These pointers are used only for comparison to determine if a 
// device is input or output - do not dereference them. They
// are integers because treating them as pointers causes segfault.
static int mmfakealsa_read_pcm = 0;
static int mmfakealsa_write_pcm = 0;


/*************************************************** 
snd_pcm_info_t functions set/get general info about a 
PCM device. Since multiple soundcards aren't supported 
and mplayer, xmms and flash don't seem to need much of 
this, it's a minimal implementation.
***************************************************/

int snd_pcm_info(snd_pcm_t *pcm, snd_pcm_info_t *info)
{
	mmfakealsa_dbg("snd_pcm_info()\n");
	return 0;
}

size_t snd_pcm_info_sizeof(void)
{
	mmfakealsa_dbg("snd_pcm_info_sizeof()\n");
	return sizeof(snd_pcm_info_t);
}

void snd_pcm_info_set_subdevice(snd_pcm_info_t *obj, unsigned int val)
{
	mmfakealsa_dbg("snd_pcm_info_set_subdevice(%d)\n", val);
}

const char *snd_pcm_info_get_name(const snd_pcm_info_t *obj)
{
	return "default";
}

const char *snd_pcm_info_get_id(const snd_pcm_info_t *obj)
{
	return "default";
}


int snd_pcm_info_get_card(const snd_pcm_info_t *obj)
{
	return 0;
}

unsigned int snd_pcm_info_get_device(const snd_pcm_info_t *obj)
{
	return 0;
}

unsigned int snd_pcm_info_get_subdevice(const snd_pcm_info_t *obj)
{
	return 0;
}

void snd_pcm_info_set_stream(snd_pcm_info_t *obj, snd_pcm_stream_t val)
{
}

void snd_pcm_info_set_device(snd_pcm_info_t *obj, unsigned int val)
{
}


/***************************************************
ALSA provides the capability for asynchronous operation 
with the following callback/interrupt mechanism. "period" 
is the ALSA term for time between interrupts. Callbacks
are implemented in mmfakealsa using setitimer() and
SIGALRM handling.
***************************************************/ 

typedef struct _snd_async_handler snd_async_handler_t;
typedef void (*snd_async_callback_t)(snd_async_handler_t *handler);
typedef void (*sighandler_t)(int);

struct _snd_async_handler
{
	int dummy;
};

static void *mmfakealsa_async_private = NULL;
static sighandler_t mmfakealsa_old_alarm_handler = NULL;
static snd_pcm_t *mmfakealsa_async_pcm = NULL;
static snd_async_handler_t *mmfakealsa_async_handler = NULL;
static snd_async_callback_t mmfakealsa_async_callback = NULL;

static void mmfakealsa_alarm_handler(int signal)
{
	//mmfakealsa_dbg("mmfakealsa_alarm_handler(%d, %d frame period)\n", signal, mmfakealsa_period_frames);

	if (mmfakealsa_async_callback)
		(*mmfakealsa_async_callback)(mmfakealsa_async_handler);
}

static int mmfakealsa_stop_alarm()
{
	if (mmfakealsa_old_alarm_handler)
	{
		signal(SIGALRM, mmfakealsa_old_alarm_handler);
		mmfakealsa_old_alarm_handler = NULL;
	}

	struct itimerval timer;
	timer.it_value.tv_sec = 0;
	timer.it_value.tv_usec = 0;
	timer.it_interval = timer.it_value;

	if (setitimer(ITIMER_REAL, &timer, NULL) < 0)
	{
		int status = -errno;
		fprintf(stderr, "setitimer() failure: %s\n", strerror(-status));
		return status;
	}

	return 0;
}

static int mmfakealsa_start_alarm(int frames)
{
	if (frames <= 0)
		return mmfakealsa_stop_alarm();

	if (!mmfakealsa_old_alarm_handler)
		mmfakealsa_old_alarm_handler = signal(SIGALRM, mmfakealsa_alarm_handler);

	int usec = (int) (frames * 1000000.0 / (float) mmfakealsa_sample_rate);
	struct itimerval timer;
	timer.it_value.tv_sec = usec / 1000000;
	timer.it_value.tv_usec = usec % 1000000;
	timer.it_interval = timer.it_value;

	if (setitimer(ITIMER_REAL, &timer, NULL) < 0)
	{
		int status = -errno;
		fprintf(stderr, "setitimer() failure: %s\n", strerror(-status));
		return status;
	}

	return 0;
}

int snd_async_add_pcm_handler(snd_async_handler_t **handler, snd_pcm_t *pcm,
                              snd_async_callback_t callback, void *private_data)
{
	if (!callback)
		return -EINVAL;

	handler = malloc(sizeof(snd_async_handler_t));
	mmfakealsa_async_pcm = pcm;
	mmfakealsa_async_handler = *handler;
	mmfakealsa_async_callback = callback;
	mmfakealsa_async_private = private_data;
	
	mmfakealsa_dbg("snd_async_add_pcm_handler()\n");

	return 0;
}

int snd_async_del_handler(snd_async_handler_t *handler)
{
	mmfakealsa_dbg("snd_async_del_pcm_handler()\n");

	if (!handler)
		return -EINVAL;

	free(handler);

	return 0;
}

snd_pcm_t *snd_async_handler_get_pcm(snd_async_handler_t *handler)
{
	//mmfakealsa_dbg("snd_async_handler_get_pcm()\n");
	return mmfakealsa_async_pcm;
}

void *snd_async_handler_get_callback_private(snd_async_handler_t *ahandler)
{
	//mmfakealsa_dbg("snd_async_handler_get_callback_private()\n");
	return mmfakealsa_async_private;
}


/***************************************************
ALSA devices have a life cycle with a number of different states.
With Flash (tm), this life cycle is:

	snd_pcm_open() => SND_PCM_STATE_OPEN
	snd_pcm_hw_params() => SND_PCM_STATE_SETUP
	snd_pcm_prepare() => SND_PCM_STATE_PREPARED
	snd_pcm_writei() => SND_PCM_STATE_RUNNING
	snd_pcm_drop() => SND_PCM_STATE_SETUP
	snd_pcm_close()

States for pausing, draining, overflow conditions and
underflow conditions are defined but not supported in mmfakealsa.
***************************************************/ 

int snd_pcm_open(snd_pcm_t **pcm, const char *name, snd_pcm_stream_t stream, int mode)
{
	mmfakealsa_dbg("snd_pcm_open(%s, mode %d)\n", name, mode);

	// mode: 0, SND_PCM_NONBLOCK (1), or SND_PCM_ASYNC (2)
	// stream: SND_PCM_STREAM_PLAYBACK, SND_PCM_STREAM_CAPTURE

	if ((mode != 0) && (mode != SND_PCM_NONBLOCK))
	{
		fprintf(stderr, "snd_pcm_open() invalid mode %d\n", mode);
		return -EINVAL;
	}

	// Only one soundcard device is implemented for now
	if (!strstr(name, "hw:0") && !strstr(name, "default"))
		return -EINVAL;

	if (mmfakealsa_descriptor < 0)
		mmfakealsa_descriptor = open(mmfakealsa_oss_device, O_RDWR, 0644);

	if (mmfakealsa_descriptor < 0)
	{
		int status = -errno;
		fprintf(stderr, "Failure opening OSS device %s: %s\n", mmfakealsa_oss_device, strerror(-status));
		return status;
	}

	// Do not use the pcm pointer after allocation. For some reason,
	// even just referencing the value of the pointer causes a segfault.
	// That is why it is typecast to an int here and elsewhere.
	// Yes, leaving this malloc()ed results in a small memory leak, but
	// audio applications generally don't stay running for very long.

	*pcm = (snd_pcm_t*) malloc(sizeof(snd_pcm_stream_t));

	mmfakealsa_pcm_state = SND_PCM_STATE_OPEN;
	mmfakealsa_async_private = NULL;
	mmfakealsa_async_pcm = NULL;
	mmfakealsa_async_handler = NULL;
	mmfakealsa_async_callback = NULL;

	if (stream == SND_PCM_STREAM_PLAYBACK)
		mmfakealsa_write_pcm = (int) *pcm;

	else if (stream == SND_PCM_STREAM_CAPTURE)
		mmfakealsa_read_pcm = (int) *pcm;

	mmfakealsa_dbg("snd_pcm_open(%s, %d) = %d, 0x%x\n", name, mode, mmfakealsa_descriptor, (int) *pcm);

	return 0;
}

int snd_pcm_open_lconf(snd_pcm_t **pcm, const char *name, snd_pcm_stream_t stream, int mode, snd_config_t *lconf)
{
	mmfakealsa_dbg("snd_pcm_open_lconf()\n");
	return snd_pcm_open(pcm, name, stream, mode);
}

int snd_pcm_state(snd_pcm_t *pcm)
{
	if (!pcm)
		return -EINVAL;

	//mmfakealsa_dbg("snd_pcm_state() = %d\n", mmfakealsa_pcm_state);

	return mmfakealsa_pcm_state;
}

int snd_pcm_prepare(snd_pcm_t *pcm)
{
	if (!pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_prepare()\n");

	mmfakealsa_pcm_state = SND_PCM_STATE_PREPARED;

	return 0;
}

int snd_pcm_start(snd_pcm_t *pcm)
{
	if (!pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_start()\n");

	mmfakealsa_pcm_state = SND_PCM_STATE_RUNNING;

	mmfakealsa_start_alarm(mmfakealsa_period_frames);

	return 0;
}


int snd_pcm_pause(snd_pcm_t *pcm, int enable)
{
	if (!pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_pause()\n");

	mmfakealsa_pcm_state = SND_PCM_STATE_PAUSED;

	mmfakealsa_stop_alarm();

	return 0;
}

int snd_pcm_resume(snd_pcm_t *pcm)
{
	if (!pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_resume()\n");

	mmfakealsa_pcm_state = SND_PCM_STATE_RUNNING;

	mmfakealsa_start_alarm(mmfakealsa_period_frames);

	return 0;
}

int snd_pcm_drop(snd_pcm_t *pcm)
{
	// discards all unsounded frames and stop the device

	mmfakealsa_dbg("snd_pcm_drop()\n");

	mmfakealsa_stop_alarm();

	mmfakealsa_pcm_state = SND_PCM_STATE_SETUP;

	// mmfakealsa_alarm_handler(0);

	return 0;
}

int snd_pcm_drain(snd_pcm_t *pcm)
{
	// wait for all pending samples to be played and stop the device

	mmfakealsa_dbg("snd_pcm_drain()\n");

	mmfakealsa_stop_alarm();

	mmfakealsa_pcm_state = SND_PCM_STATE_SETUP;

	return 0;
}

int snd_pcm_close(snd_pcm_t *pcm)
{
	mmfakealsa_dbg("snd_pcm_close(0x%x %d)\n", (int)pcm, mmfakealsa_descriptor);

	mmfakealsa_stop_alarm();

	if (mmfakealsa_descriptor >= 0)
		close(mmfakealsa_descriptor);

	mmfakealsa_descriptor = -1;
	mmfakealsa_async_private = NULL;
	mmfakealsa_async_pcm = NULL;
	mmfakealsa_async_handler = NULL;
	mmfakealsa_async_callback = NULL;
	mmfakealsa_pcm_state = SND_PCM_STATE_DISCONNECTED;

	if (mmfakealsa_write_pcm == (int) pcm)
		mmfakealsa_write_pcm = 0;
		
	if (mmfakealsa_read_pcm == (int) pcm)
		mmfakealsa_read_pcm = 0;
	
	return 0;
}


/*************************************************** 
snd_pcm_writei() and snd_pcm_readi() write/read interleaved 
samples to the device, returning the number of frames
actually written.
***************************************************/

snd_pcm_sframes_t snd_pcm_avail_update(snd_pcm_t *);

snd_pcm_sframes_t snd_pcm_writei(snd_pcm_t *pcm, const void *buffer, snd_pcm_uframes_t frames)
{
	if (!pcm || !buffer || (frames <= 0) || (mmfakealsa_descriptor < 0) || (mmfakealsa_channels < 1))
		return -EINVAL;

	// Flash 9 constantly writes to the device even when waiting for free buffer space.
	// Therefore, if there isn't enough free space to write all requested frames,
	// return 0 indicating that no frames were written. This probably should
	int available = snd_pcm_avail_update(pcm);
	if (frames > available)
		frames = available;

	int bytes = frames * mmfakealsa_channels * (mmfakealsa_bits_per_sample / 8);

	//mmfakealsa_dbg("snd_pcm_writei(%lu frames = %d bytes)\n", frames, bytes);

	// If a device is prepared, a snd_pcm_writei() call can start it running - snd_pcm_start() not needed.
	if (mmfakealsa_pcm_state != SND_PCM_STATE_RUNNING)
	{
		mmfakealsa_start_alarm(mmfakealsa_period_frames);
		mmfakealsa_pcm_state = SND_PCM_STATE_RUNNING;
	}

	if (write(mmfakealsa_descriptor, buffer, bytes) < 0)
	{
		int status = -errno;
		fprintf(stderr, "snd_pcm_writei(%lu) failure: %s\n", frames, strerror(-status));
		return status;
	}
	else
		return frames;
}

snd_pcm_sframes_t snd_pcm_readi(snd_pcm_t *pcm, void *buffer, snd_pcm_uframes_t frames)
{
	if (!pcm || !buffer || (frames <= 0) || (mmfakealsa_descriptor < 0) || (mmfakealsa_channels < 1))
		return -EINVAL;

	int bytes = frames * mmfakealsa_channels * (mmfakealsa_bits_per_sample / 8);

	mmfakealsa_pcm_state = SND_PCM_STATE_RUNNING;

	//mmfakealsa_dbg("snd_pcm_readi(%d frames = %d bytes)\n", frames, bytes);

	if (read(mmfakealsa_descriptor, buffer, bytes) < 0)
		return -errno;
	else
		return frames;
}


/*************************************************** 
snd_pcm_delay() returns time delay in frames between 
the current current position and the actual audio 
output (i.e. number of frames waiting in the output buffer)
***************************************************/

int snd_pcm_delay(snd_pcm_t *pcm, snd_pcm_sframes_t *delayp)
{
	if (!delayp || (mmfakealsa_channels <= 0) || (mmfakealsa_bits_per_sample <= 0))
		return -EINVAL; // prevent div by zero error

	audio_buf_info info;
	if (ioctl(mmfakealsa_descriptor, SNDCTL_DSP_GETOSPACE, &info) < 0)
	{
		int status = -errno;
		fprintf(stderr, "ioctl(SNDCTL_DSP_GETOSPACE) failure: %s\n", strerror(-status));
		return status;
	}

	int frames = info.bytes / mmfakealsa_channels / (mmfakealsa_bits_per_sample / 8);
	int total_frames = info.fragstotal / mmfakealsa_channels / (mmfakealsa_bits_per_sample / 8);

	*delayp = total_frames - frames;

	//mmfakealsa_dbg("snd_pcm_delay() = %lu\n", *delayp);

	return 0;
}

/*************************************************** 
snd_pcm_avail_update() gets the number of frames that
can be written to the device (i.e. free space left
in the buffer)
***************************************************/

snd_pcm_sframes_t snd_pcm_avail_update(snd_pcm_t *pcm)
{
	if ((mmfakealsa_channels <= 0) || (mmfakealsa_bits_per_sample <= 0))
		return -EINVAL; // prevent div by zero error

	audio_buf_info info;
	if (ioctl(mmfakealsa_descriptor, SNDCTL_DSP_GETOSPACE, &info) < 0)
	{
		int status = -errno;
		fprintf(stderr, "ioctl(SNDCTL_DSP_GETOSPACE) failure: %s\n", strerror(-status));
		return status;
	}

	int frames = info.bytes / mmfakealsa_channels / (mmfakealsa_bits_per_sample / 8);
	int total_frames = info.fragstotal / mmfakealsa_channels / (mmfakealsa_bits_per_sample / 8);

	// kludge for Flash 9
	if (frames <= 66) 
		frames = 0;

	// kludge for JACK
	if (frames > mmfakealsa_buffer_frames)
		frames = mmfakealsa_buffer_frames;

	//mmfakealsa_dbg("snd_pcm_avail_update() = %d of %d frames (%d bytes)\n", frames, total_frames, info.bytes);

	return frames;
}


/***************************************************
Nonblocking writes are not reliably available with OSS.
But since mplayer, xmms and flash seem to work fine
with blocking writes, this is not a problem.
***************************************************/

int snd_pcm_nonblock(snd_pcm_t *pcm, int nonblock)
{
	mmfakealsa_dbg("snd_pcm_nonblock()\n");
	return 0;
}


int snd_pcm_wait(snd_pcm_t *pcm, int timeout)
{
	mmfakealsa_dbg("snd_pcm_wait()\n");
	return 0;
}

int snd_pcm_link(snd_pcm_t *pcm1, snd_pcm_t *pcm2)
{
	// Not implemented yet
	return -EINVAL;
}

int snd_pcm_unlink(snd_pcm_t *pcm)
{
	// Not implemented yet
	return -EINVAL;
}



/*************************************************** 
snd_pcm_forward() advances application frame position.
Unsupported for now.
***************************************************/

snd_pcm_sframes_t snd_pcm_forward(snd_pcm_t *pcm, snd_pcm_uframes_t frames)
{
	mmfakealsa_dbg("snd_pcm_forward(%ld)\n", frames);

	return 0;
}



/***************************************************
snd_pcm_dump() prints a diagnostic listing?
***************************************************/ 

int snd_pcm_dump(snd_pcm_t *pcm, snd_output_t *out)
{
	return 0;
}
	

/***************************************************
snd_pcm_format_width() returns the number of bits for a given audio format
***************************************************/ 

int snd_pcm_format_width(snd_pcm_format_t format)
{
        switch (format) 
	{
        	case SND_PCM_FORMAT_S8:
	        case SND_PCM_FORMAT_U8:
        	        return 8;
	        case SND_PCM_FORMAT_S16_LE:
        	case SND_PCM_FORMAT_S16_BE:
	        case SND_PCM_FORMAT_U16_LE:
        	case SND_PCM_FORMAT_U16_BE:
                	return 16;
	        case SND_PCM_FORMAT_S18_3LE:
        	case SND_PCM_FORMAT_S18_3BE:
	        case SND_PCM_FORMAT_U18_3LE:
        	case SND_PCM_FORMAT_U18_3BE:
                	return 18;
	        case SND_PCM_FORMAT_S20_3LE:
        	case SND_PCM_FORMAT_S20_3BE:
	        case SND_PCM_FORMAT_U20_3LE:
        	case SND_PCM_FORMAT_U20_3BE:
                	return 20;
	        case SND_PCM_FORMAT_S24_LE:
        	case SND_PCM_FORMAT_S24_BE:
	        case SND_PCM_FORMAT_U24_LE:
        	case SND_PCM_FORMAT_U24_BE:
	        case SND_PCM_FORMAT_S24_3LE:
        	case SND_PCM_FORMAT_S24_3BE:
	        case SND_PCM_FORMAT_U24_3LE:
        	case SND_PCM_FORMAT_U24_3BE:
                	return 24;
	        case SND_PCM_FORMAT_S32_LE:
        	case SND_PCM_FORMAT_S32_BE:
	        case SND_PCM_FORMAT_U32_LE:
        	case SND_PCM_FORMAT_U32_BE:
	        case SND_PCM_FORMAT_FLOAT_LE:
        	case SND_PCM_FORMAT_FLOAT_BE:
                	return 32;
	        case SND_PCM_FORMAT_FLOAT64_LE:
        	case SND_PCM_FORMAT_FLOAT64_BE:
                	return 64;
	        case SND_PCM_FORMAT_IEC958_SUBFRAME_LE:
        	case SND_PCM_FORMAT_IEC958_SUBFRAME_BE:
                	return 32;
	        case SND_PCM_FORMAT_MU_LAW:
        	case SND_PCM_FORMAT_A_LAW:
                	return 8;
	        case SND_PCM_FORMAT_IMA_ADPCM:
        	        return 4;
        }

	return -EINVAL;
}

int snd_pcm_format_physical_width(snd_pcm_format_t format)
{
	// Why would "physical width" ever be different from format bit definition?
	return snd_pcm_format_width(format);
}

const char *snd_pcm_format_name(snd_pcm_format_t format)
{
	static char name[64];
	sprintf(name, "Format %d", format);
	return name;
}

const char *snd_pcm_format_description(snd_pcm_format_t format)
{
	static char name[64];
	sprintf(name, "Format %d", format);
	return name;
}


const char *snd_strerror(int errnum)
{
	static char name[64];
	sprintf(name, "Error %d", errnum);
	return name;
}


/*************************************************** 
The following convert between frame, sample and
byte sizes based on the current configuration.
***************************************************/
 
ssize_t snd_pcm_frames_to_bytes(snd_pcm_t *pcm, snd_pcm_sframes_t frames)
{
	if (!pcm)
		return -EINVAL;

	int bytes = frames * mmfakealsa_channels * (mmfakealsa_bits_per_sample / 8);

	//mmfakealsa_dbg("snd_pcm_frames_to_bytes(%d) = %d\n", frames, bytes);

	return bytes;
}

ssize_t snd_pcm_samples_to_bytes(snd_pcm_t *pcm, long samples)
{
	if (!pcm)
		return -EINVAL;

	//mmfakealsa_dbg("snd_pcm_samples_to_bytes()\n");

	// This is samples, not frames, so channel count is ignored

	return samples * mmfakealsa_bits_per_sample / 8;
}

snd_pcm_sframes_t snd_pcm_bytes_to_frames(snd_pcm_t *pcm, ssize_t bytes)
{
	if (!pcm || (mmfakealsa_channels <= 0) || (mmfakealsa_bits_per_sample <= 0))
		return -EINVAL;

	int frames = bytes / mmfakealsa_channels / (mmfakealsa_bits_per_sample / 8);

	//mmfakealsa_dbg("snd_pcm_bytes_to_frames(%d) = %d\n", bytes, frames);

	return frames;
}


/*************************************************** 
snd_pcm_status_t in ALSA is a structure that can contain
full status information. I guess it's there to permit
taking an atomic snapshot of the device status. However,
this snapshot capabilitity doesn't seem necessary for 
mplayer, xmms or Flash, so the current state is used.
***************************************************/

typedef snd_pcm_t snd_pcm_status_t;
typedef struct timeval snd_timestamp_t;

int snd_pcm_status(snd_pcm_t *pcm, snd_pcm_status_t *status)
{
	if (!pcm || !status)
		return -EINVAL;

	//mmfakealsa_dbg("snd_pcm_status()\n");

	return 0;
}

int snd_pcm_status_malloc(snd_pcm_status_t **ptr)
{
	if (!ptr)
		return -EINVAL;

        *ptr = calloc(1, sizeof(snd_pcm_status_t));
        if (!*ptr)
                return -ENOMEM;

        return 0;
}

snd_pcm_state_t snd_pcm_status_get_state(const snd_pcm_status_t *obj)
{
	if (!obj)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_status_get_state()\n");

	return mmfakealsa_pcm_state;
}

size_t snd_pcm_status_sizeof()
{
	//mmfakealsa_dbg("snd_pcm_status_sizeof()\n");
        return sizeof(snd_pcm_status_t);
}

int snd_pcm_status_dump(snd_pcm_status_t *status, snd_output_t *out)
{
	return -EINVAL;
}

void snd_pcm_status_get_trigger_tstamp(const snd_pcm_status_t *obj, snd_timestamp_t *ptr)
{
	if (ptr)
		gettimeofday(ptr, NULL);

	mmfakealsa_dbg("snd_pcm_status_get_trigger_tstamp()\n");
}

snd_pcm_uframes_t snd_pcm_status_get_avail(const snd_pcm_status_t *obj)
{
	// Get frames that can be read/written
	return snd_pcm_avail_update(NULL);
}



/*************************************************** 
snd_pcm_hw_params is a nasty, kludgy structure used 
for keeping track of device configuration parameters.
Because of alloca() segfault problems actually using 
the structure, the global configuration values are
set instead with snd_pcm_hw_params() ioctl()ing
the values to the device.
***************************************************/ 

typedef struct snd_pcm_hw_params snd_pcm_hw_params_t;

struct snd_pcm_hw_params
{
	int dummy;
};

int snd_pcm_hw_params(snd_pcm_t *pcm, snd_pcm_hw_params_t *params)
{
	if (!pcm || !params || (mmfakealsa_descriptor < 0))
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params()\n");

	int status = 0;
	if (ioctl(mmfakealsa_descriptor, SNDCTL_DSP_SETFMT, &mmfakealsa_oss_format) < 0)
	{
		status = -errno;
		fprintf(stderr, "Failure setting audio device format: %s\n", strerror(errno));
	}

	else if (ioctl(mmfakealsa_descriptor, SNDCTL_DSP_CHANNELS, &mmfakealsa_channels) < 0)
	{
		status = -errno;
		fprintf(stderr, "Failure setting audio device channels: %s\n", strerror(errno));
	}

	else if (ioctl(mmfakealsa_descriptor, SNDCTL_DSP_SPEED, &mmfakealsa_sample_rate) < 0)
	{
		status = -errno;
		fprintf(stderr, "Failure setting audio device sample rate: %s\n", strerror(errno));
	}

	mmfakealsa_pcm_state = SND_PCM_STATE_SETUP;

	return status;
}


size_t snd_pcm_hw_params_sizeof(void)
{
	int size = 604;
	mmfakealsa_dbg("snd_pcm_hw_params_sizeof() = %d\n", size);
	//return sizeof(struct snd_pcm_hw_params);
	return size;
}

int snd_pcm_hw_params_malloc(snd_pcm_hw_params_t **ptr)
{
	if (!ptr)
		return -EINVAL;
	*ptr = calloc(1, snd_pcm_hw_params_sizeof());
	return 0;
}

void snd_pcm_hw_params_free(snd_pcm_hw_params_t *obj)
{
	if (obj)
		free(obj);
}

int snd_pcm_hw_params_any(snd_pcm_t *handle, snd_pcm_hw_params_t *params)
{
	// Requests configuration of all parameters - sure, why not
	mmfakealsa_dbg("snd_pcm_hw_params_any()\n");

	return 0;
}

int snd_pcm_hw_params_set_rate_resample(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int val)
{
	// Requests hardware resampling - unsupported by OSS
	return 0;
}

int snd_pcm_hw_params_set_access(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_access_t access)
{
	if ((access != SND_PCM_ACCESS_RW_INTERLEAVED) && (access != SND_PCM_ACCESS_MMAP_INTERLEAVED))
		return -EINVAL;

	/**
	SND_PCM_ACCESS_MMAP_NONINTERLEAVED,
	SND_PCM_ACCESS_MMAP_COMPLEX,
	SND_PCM_ACCESS_RW_NONINTERLEAVED
	**/

	return 0;
}

int snd_pcm_hw_params_test_format(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_format_t format)
{
	// Determine if format is available from device
	if ((format != SND_PCM_FORMAT_U8) && (format != SND_PCM_FORMAT_S16_LE))
		return -EINVAL;
	return 0;
}

int snd_pcm_hw_params_set_format(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_format_t format)
{
	if (!pcm || !params)
		return -EINVAL;

	//For now, this only supports little-endian (Intel) 8 and 16 formats.

	mmfakealsa_dbg("snd_pcm_hw_params_set_format(%d)\n", format);

	if (format == SND_PCM_FORMAT_U8)
	{
		mmfakealsa_alsa_format = format;
		mmfakealsa_oss_format = AFMT_U8;
		mmfakealsa_bits_per_sample = 8;
	}

	else if (format == SND_PCM_FORMAT_S16_LE)
	{
		mmfakealsa_alsa_format = format;
		mmfakealsa_oss_format = AFMT_S16_LE; 
		mmfakealsa_bits_per_sample = 16;
	}

	else
		return -EINVAL;

	return 0;
}

int snd_pcm_hw_params_set_channels(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int channels)
{
	mmfakealsa_dbg("snd_pcm_hw_params_set_channels(%d)\n", channels);

	if (!params || !pcm)
		return -EINVAL;

	if ((channels < 1) || (channels > 2))
		return -EINVAL;

	mmfakealsa_channels = channels;

	return 0;
}

int snd_pcm_hw_params_set_channels_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val)
{
	if (!params || !val || !pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_channels_near(%d)\n", *val);

	if (*val < 1)
		*val = 1;
	if (*val > 2)
		*val = 2;

	mmfakealsa_channels = *val;

	return 0;
}

int snd_pcm_hw_params_set_rate_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	if (!val)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_rate_near(%d)\n", *val);

	if (*val < 4410)
		*val = 4410;

	if (*val > 48000)
		*val = 48000;

	mmfakealsa_sample_rate = *val;

	return 0;
}

int snd_pcm_hw_params_set_buffer_time_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	if (!params || !val || !pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_buffer_time_near(%dus)\n", *val);

	if (*val < 10000)
		*val = 10000;

	if (*val > 1000000)
		*val = 1000000;

	mmfakealsa_buffer_frames = (int)(*val * (float) mmfakealsa_sample_rate / 1000000.0);

	return 0;
}

int snd_pcm_hw_params_set_buffer_size_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_uframes_t *val)
{
	if (!params || !val || !pcm || (mmfakealsa_sample_rate <= 0))
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_buffer_size_near(%d frames)\n", *val);

	mmfakealsa_buffer_frames = *val;

	return 0;
}

int snd_pcm_hw_params_set_buffer_size(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_uframes_t val)
{
	if (!val || (mmfakealsa_sample_rate <= 0))
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_buffer_size(%d frames)\n", val);

	mmfakealsa_buffer_frames = val;

	return 0;
}


int snd_pcm_hw_params_set_period_time_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	if (!params || !val || !pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_period_time_near(%dus)\n", *val);
	mmfakealsa_period_frames = (int)(*val * (float) mmfakealsa_sample_rate / 1000000.0);

	return 0;
}

int snd_pcm_hw_params_set_period_size(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_uframes_t val, int dir)
{
	mmfakealsa_dbg("snd_pcm_hw_params_set_period_size(%d frames)\n", val);
	if (mmfakealsa_sample_rate <= 0)
		return -EINVAL;
	mmfakealsa_period_frames = val;
	return 0;
}


int snd_pcm_hw_params_set_period_size_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, snd_pcm_uframes_t *val, int *dir)
{
	if (!params || !val || !pcm)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_hw_params_set_period_size_near(%d frames)\n", *val);

	mmfakealsa_period_frames = *val;

	return 0;
}


int snd_pcm_hw_params_set_periods_integer(snd_pcm_t *pcm, snd_pcm_hw_params_t *params)
{
	return 0;
}

int snd_pcm_hw_params_set_periods_min(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	return 0;
}

int snd_pcm_hw_params_set_periods_max(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	return 0;
}

int snd_pcm_hw_params_set_periods_near(snd_pcm_t *pcm, snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	// val is periods per buffer
	return 0;
}

/**
4/25/07 - Defining this causes segfault in mplayer for some reason???

int snd_pcm_hw_params_get_channels(const snd_pcm_hw_params_t *params, unsigned int *val)
{
	if (!params || !val)
		return -EINVAL;

	*val = mmfakealsa_channels;

	return 0;
}
**/

int snd_pcm_hw_params_get_channels_max(const snd_pcm_hw_params_t *params, unsigned int *val)
{
	if (!val)
		return -EINVAL;
	*val = 2;
	return 0;
}


int snd_pcm_hw_params_get_buffer_size(const snd_pcm_hw_params_t *params, snd_pcm_uframes_t *val)
{
	if (!params || !val)
		return -EINVAL;

	*val = mmfakealsa_buffer_frames;

	mmfakealsa_dbg("snd_pcm_hw_params_get_buffer_size()\n");

	return 0;
}

int snd_pcm_hw_params_get_format(const snd_pcm_hw_params_t *params, snd_pcm_format_t *val)
{
	if (!val)
		return -EINVAL;

	*val = mmfakealsa_alsa_format;

	return 0;
}

int snd_pcm_hw_params_get_period_size(const snd_pcm_hw_params_t *params, snd_pcm_uframes_t *val, int *dir)
{
	if (!params || !val)
		return -EINVAL;

	*val = mmfakealsa_period_frames;

	mmfakealsa_dbg("snd_pcm_hw_params_get_period_size() = %d frames\n", *val);

	return 0;
}

int snd_pcm_hw_params_get_rate(const snd_pcm_hw_params_t *params, unsigned int *val, int *dir)
{
	if (!val)
		return -EINVAL;

	*val = mmfakealsa_sample_rate;

	return 0;
}

int snd_pcm_hw_params_get_access(const snd_pcm_hw_params_t *params, snd_pcm_access_t *_access)
{
	if (!_access)
		return -EINVAL;

	*_access = SND_PCM_ACCESS_MMAP_INTERLEAVED;

	return 0;
}

int snd_pcm_hw_params_can_pause(const snd_pcm_hw_params_t *params)
{
	return 1;
}

int snd_pcm_hw_params_dump(snd_pcm_hw_params_t *params, snd_output_t *out)
{
	return 0;
}


/*************************************************** 
snd_pcm_sw_params is a nasty, kludgy structure used 
for keeping track of device configuration parameters.
Software parameters are not implemented so the following 
functions do nothing.
***************************************************/

typedef struct snd_pcm_sw_params snd_pcm_sw_params_t;

struct snd_pcm_sw_params 
{
	int dummy;
};

size_t snd_pcm_sw_params_sizeof(void)
{
	int size = 104;
	mmfakealsa_dbg("snd_pcm_sw_params_sizeof() = %d\n", size);

	//return sizeof(snd_pcm_sw_params_t);
	return size;
}

int snd_pcm_sw_params_malloc(snd_pcm_sw_params_t **ptr)
{
	if (!ptr)
		return -EINVAL;
	*ptr = calloc(1, snd_pcm_sw_params_sizeof());
	return 0;
}

void snd_pcm_sw_params_free(snd_pcm_sw_params_t *obj)
{
	if (obj)
		free(obj);
}


int snd_pcm_sw_params_current(snd_pcm_t *pcm, snd_pcm_sw_params_t *params)
{
	return 0;
}

int snd_pcm_sw_params_set_start_threshold(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, snd_pcm_uframes_t val)
{
	return 0;
}

int snd_pcm_sw_params_set_avail_min(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, snd_pcm_uframes_t val)
{
	return 0;
}

int snd_pcm_sw_params_set_xfer_align(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, snd_pcm_uframes_t val)
{
	return 0;
}

int snd_pcm_sw_params_set_stop_threshold(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, snd_pcm_uframes_t val)
{
	return 0;
}

int snd_pcm_sw_params_set_silence_size(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, snd_pcm_uframes_t val)
{
	return 0;
}

int snd_pcm_sw_params_set_silence_threshold(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, snd_pcm_uframes_t val)
{
	return 0;
}

int snd_pcm_sw_params_set_sleep_min(snd_pcm_t *pcm, snd_pcm_sw_params_t *params, unsigned int val)
{
	return 0;
}

int snd_pcm_sw_params_get_boundary(const snd_pcm_sw_params_t *params, snd_pcm_uframes_t *val)
{
	if (!val)
		return -EINVAL;

	mmfakealsa_dbg("snd_pcm_sw_params_get_boundary()\n");

	*val = 0x7fffffff; // arbitrary value from mplayer
	
	return 0;
}

int snd_pcm_sw_params(snd_pcm_t *pcm, snd_pcm_sw_params_t *params)
{
	mmfakealsa_dbg("snd_pcm_sw_params()\n");
	return 0;
}

int snd_pcm_sw_params_dump(snd_pcm_sw_params_t *params, snd_output_t *out)
{
	return 0;
}
	

/*************************************************** 
Mixers are unsupported for now but the functions 
need to be defined so the load of libasound will
not fail with programs that use the mixer.
***************************************************/


typedef struct _snd_mixer snd_mixer_t;
typedef struct _snd_mixer_elem snd_mixer_elem_t;
typedef struct _snd_mixer_class snd_mixer_class_t;
typedef struct _snd_mixer_selem_id snd_mixer_selem_id_t;

struct _snd_mixer_selem_id
{
	int dummy;
};

typedef enum _snd_mixer_selem_channel_id 
{
        SND_MIXER_SCHN_UNKNOWN = -1,
        SND_MIXER_SCHN_FRONT_LEFT = 0,
        SND_MIXER_SCHN_FRONT_RIGHT,
        SND_MIXER_SCHN_REAR_LEFT,
        SND_MIXER_SCHN_REAR_RIGHT,
        SND_MIXER_SCHN_FRONT_CENTER,
        SND_MIXER_SCHN_WOOFER,
        SND_MIXER_SCHN_SIDE_LEFT,
        SND_MIXER_SCHN_SIDE_RIGHT,
        SND_MIXER_SCHN_REAR_CENTER,
        SND_MIXER_SCHN_LAST = 31,
        SND_MIXER_SCHN_MONO = SND_MIXER_SCHN_FRONT_LEFT
} snd_mixer_selem_channel_id_t;

enum snd_mixer_selem_regopt_abstract 
{
        SND_MIXER_SABSTRACT_NONE = 0,	/** no abstraction - try use all universal controls from driver */
        SND_MIXER_SABSTRACT_BASIC,	/** basic abstraction - Master,PCM,CD,Aux,Record-Gain etc. */
};

struct snd_mixer_selem_regopt 
{
        int ver;			/** structure version */
        enum snd_mixer_selem_regopt_abstract abstract; /** v1: abstract layer selection */
        const char *device;		/** v1: device name (must be NULL when playback_pcm or capture_pcm != NULL) */
        snd_pcm_t *playback_pcm;	/** v1: playback PCM connected to mixer device (NULL == none) */
        snd_pcm_t *capture_pcm;		/** v1: capture PCM connected to mixer device (NULL == none) */
};

int snd_mixer_open(snd_mixer_t **mixer, int mode)
{
	return -EINVAL;
}

int snd_mixer_close(snd_mixer_t *mixer)
{
	return -EINVAL;
}

size_t snd_mixer_selem_id_sizeof(void)
{
	return sizeof(snd_mixer_selem_id_t);
}

void snd_mixer_selem_id_set_name(snd_mixer_selem_id_t *obj, const char *val)
{
	mmfakealsa_dbg("snd_mixer_selem_id_set_name(%s)\n", val);
}

void snd_mixer_selem_id_set_index(snd_mixer_selem_id_t *obj, unsigned int val)
{
	mmfakealsa_dbg("snd_mixer_selem_id_set_index(%d)\n", val);
}

const char *snd_mixer_selem_id_get_name(const snd_mixer_selem_id_t *obj)
{
	return NULL;
}

unsigned int snd_mixer_selem_id_get_index(const snd_mixer_selem_id_t *obj)
{
	return 0;
}


snd_mixer_elem_t *snd_mixer_find_selem(snd_mixer_t *mixer, const snd_mixer_selem_id_t *id)
{
	return NULL;
}

snd_mixer_elem_t *snd_mixer_first_elem(snd_mixer_t *mixer)
{
	return NULL;
}

snd_mixer_elem_t *snd_mixer_elem_next(snd_mixer_elem_t *elem)
{
	return NULL;
}

int snd_mixer_selem_register(snd_mixer_t *mixer, struct snd_mixer_selem_regopt *options, snd_mixer_class_t **classp)
{
	return -EINVAL;
}

const char *snd_mixer_selem_get_name(snd_mixer_elem_t *elem)
{
	return "default";
}

int snd_mixer_handle_events(snd_mixer_t *mixer)
{
	return -EINVAL;
}

int snd_mixer_load(snd_mixer_t *mixer)
{
	return -EINVAL;
}

int snd_mixer_attach(snd_mixer_t *mixer, const char *name)
{
	mmfakealsa_dbg("snd_mixer_attach(%s)\n", name);
	return -EINVAL;
}

int snd_mixer_selem_set_playback_switch(snd_mixer_elem_t *elem, snd_mixer_selem_channel_id_t channel, int value)
{
	mmfakealsa_dbg("snd_mixer_selem_set_playback_switch(channel %d, value %d)\n", channel, value);
	return 0;
}


int snd_mixer_selem_set_playback_volume(snd_mixer_elem_t *elem, snd_mixer_selem_channel_id_t channel, long value)
{
	mmfakealsa_dbg("snd_mixer_selem_set_playback_volume(channel %d, volume %ld)\n", channel, value);
	return 0;
}

int snd_mixer_selem_get_playback_volume(snd_mixer_elem_t *elem, snd_mixer_selem_channel_id_t channel, long *value)
{
	return -EINVAL;
}

int snd_mixer_selem_set_playback_volume_range(snd_mixer_elem_t *elem, long min, long max)
{
	return -EINVAL;
}

int snd_mixer_selem_get_playback_volume_range(snd_mixer_elem_t *elem, long *min, long *max)
{
	if (!min || !max)
		return -EINVAL;

	// Use percent for volume settings
	*min = 0;
	*max = 100;

	return 0;
}

int snd_mixer_selem_is_active(snd_mixer_elem_t *elem)
{
	return 0;
}

int snd_mixer_selem_has_playback_volume(snd_mixer_elem_t *elem)
{
	return 0;
}

int snd_mixer_selem_has_playback_switch(snd_mixer_elem_t *elem)
{
	return 0;
}

int snd_mixer_selem_has_playback_switch_joined(snd_mixer_elem_t *elem)
{
	return 0;
}




/*************************************************** 
Polling is unsupported for now but the functions 
need to be defined so the load of libasound will
not fail with programs that use polling.
***************************************************/

int snd_pcm_poll_descriptors_revents(snd_pcm_t *pcm, struct pollfd *pfds, unsigned int nfds, unsigned short *revents)
{
	if (!pcm || !revents)
		return -EINVAL;

	*revents = 0;
	return -EINVAL;
}

int snd_pcm_poll_descriptors(snd_pcm_t *pcm, struct pollfd *pfds, unsigned int space)
{
	// mmfakealsa_dbg("snd_pcm_poll_descriptors()\n");

	pfds->fd = mmfakealsa_descriptor;
	pfds->events = POLLIN | POLLOUT; // must be set, otherwise poll() times out
	pfds->revents = 0;

	return 0;
}

int snd_pcm_poll_descriptors_count(snd_pcm_t *pcm)
{
	mmfakealsa_dbg("snd_pcm_poll_descriptors_count()\n");

	return 1;
}


/*************************************************** 
Memory mapping can be used to reduce the number of
buffer copies when exchanging sample data. Memory
mapping is unsupported for now but the functions 
need to be defined so the load of libasound will
not fail with programs that use polling.
***************************************************/

typedef struct _snd_pcm_channel_area 
{
	void *addr;		/** base address of channel samples */
	unsigned int first;	/** offset to first sample in bits */
	unsigned int step;	/** samples distance in bits */
} snd_pcm_channel_area_t;

static char *mmfakealsa_mmap_buffer = NULL;
static int mmfakealsa_mmap_buffer_size = 0;
static snd_pcm_channel_area_t mmfakealsa_mmap_buffer_info[2]; // left/right

int snd_pcm_mmap_begin(snd_pcm_t *pcm, const snd_pcm_channel_area_t **areas, 
		       snd_pcm_uframes_t *offset, snd_pcm_uframes_t *frames)
{
	if (!areas || !offset || !frames)
		return -EINVAL;

	//mmfakealsa_dbg("snd_pcm_mmap_begin(pcm 0x%x, %d offset, %d frames)\n", (int) pcm, (int) *offset, (int) *frames);

	if (!mmfakealsa_mmap_buffer)
	{
		mmfakealsa_mmap_buffer_size = 2 * mmfakealsa_buffer_frames * 
			mmfakealsa_channels * (mmfakealsa_bits_per_sample / 8);
		mmfakealsa_mmap_buffer = calloc(mmfakealsa_mmap_buffer_size, 1);
	}

	mmfakealsa_mmap_buffer_info[0].addr = mmfakealsa_mmap_buffer;
	mmfakealsa_mmap_buffer_info[0].first = 0;
	mmfakealsa_mmap_buffer_info[0].step = mmfakealsa_bits_per_sample * mmfakealsa_channels;
	mmfakealsa_mmap_buffer_info[1].addr = mmfakealsa_mmap_buffer;
	mmfakealsa_mmap_buffer_info[1].first = mmfakealsa_bits_per_sample;
	mmfakealsa_mmap_buffer_info[1].step = mmfakealsa_bits_per_sample * mmfakealsa_channels;
	*areas = mmfakealsa_mmap_buffer_info;

	/**
	audio_buf_info info;
	if (ioctl(mmfakealsa_descriptor, SNDCTL_DSP_GETOSPACE, &info) < 0)
	{
		int status = -errno;
		mmfakealsa_dbg("ioctl(SNDCTL_DSP_GETOSPACE) failure: %s\n", strerror(-status));
		return status;
	}

	int available_frames = info.bytes / mmfakealsa_channels / (mmfakealsa_bits_per_sample / 8);
	if (*frames > available_frames)
		*frames = available_frames;
	**/

	if (*frames > mmfakealsa_buffer_frames)
		*frames = mmfakealsa_buffer_frames;
	*offset = 0;

	//mmfakealsa_dbg("  available %d frames\n", *frames);

	return 0;
}

snd_pcm_sframes_t snd_pcm_mmap_commit(snd_pcm_t *pcm, snd_pcm_uframes_t offset, snd_pcm_uframes_t frames)
{
	//mmfakealsa_dbg("snd_pcm_mmap_commit(pcm 0x%x, %d offset, %d frames)\n", (int) pcm, (int) offset, (int) frames);

	if (mmfakealsa_write_pcm == (int) pcm)
		return snd_pcm_writei(pcm, mmfakealsa_mmap_buffer, frames);
		
	else if (mmfakealsa_read_pcm == (int) pcm)
		return snd_pcm_readi(pcm, mmfakealsa_mmap_buffer, frames);

	return frames;
}

snd_pcm_sframes_t snd_pcm_mmap_writei(snd_pcm_t *pcm, const void *buffer, snd_pcm_uframes_t size)
{
	mmfakealsa_dbg("snd_pcm_mmap_writei(%d frames)\n", size);
	return -EINVAL;
}

#if 0
Symbols in Flash 9 (tm) courtesy of vim

libasound.so
libasound.so.2
snd_pcm_open
snd_pcm_close
snd_pcm_start
snd_pcm_drop
snd_pcm_prepare
snd_pcm_resume
snd_pcm_hw_params_sizeof
snd_pcm_sw_params_sizeof
snd_pcm_status_sizeof
snd_pcm_info
snd_pcm_info_get_id
snd_pcm_info_get_name
snd_pcm_hw_params_any
snd_pcm_hw_params_set_access
snd_pcm_hw_params_set_format
snd_pcm_hw_params
snd_pcm_sw_params_current
snd_pcm_sw_params
snd_pcm_writei
snd_pcm_readi
snd_pcm_delay
snd_pcm_avail_update
snd_pcm_status_malloc
snd_pcm_status
snd_pcm_status_dump
snd_pcm_state
snd_strerror
snd_pcm_frames_to_bytes
snd_async_handler_get_pcm
snd_async_add_pcm_handler
snd_pcm_hw_params_set_channels
snd_pcm_hw_params_set_rate_near
snd_pcm_hw_params_set_buffer_time_near
snd_pcm_hw_params_get_buffer_size
snd_pcm_hw_params_set_period_time_near
snd_pcm_hw_params_get_period_size
snd_pcm_sw_params_set_start_threshold
snd_pcm_sw_params_set_stop_threshold
snd_pcm_sw_params_set_avail_min
snd_pcm_sw_params_set_xfer_align
snd_async_handler_get_callback_private

#endif

