/*****************************************************************************
   MMUSBAUDIO: USB kernel module for the Roland UA-100, UA-5, SC-8850 & UM-1

   (C) 2001-2004 by Michael Minn - see www.michaelminn.com for contact info

   printk messages are written to /var/log/messages

   See /usr/src/linux/Documentation/usb for more information about USB programming

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License as 
   published by the Free Software Foundation.


   NOTES ON AUDIO PROTOCOL FOR THE ROLAND UA-100 and UA-5 ----

   For writes, UA-100 data is four interleaved 16-bit channels (wave 1 and wave 2)
   for a total of 8 bytes per sample. UA-5 has 2 interleaved 24-bit channels
   for a total of 6 bytes per sample.

   At 44.1 Khz sampling rate, to time isosynchronous writes correctly each URB 
   contain 1 ms of data. Each frame is 44 samples except for every 10th URB, 
   which is 45 samples. 

   For the UA-100, the URBs are 360 and 352 bytes, respectively. Accordingly, the
   UA-5 URBs are 270 and 263 (246) bytes, respectively.

   For Reads, however, each frame must be sent at full length (184 bytes) or you 
   will get a nasty kernel message in /var/log/messages: 
   "usb-uhci.c: interrupt, status 2, frame# xxx"

   At 48 Khz and 96 Khz, this uneven frame size is not needed since 1ms
   equals 48 and 96 samples, respectively.

****************************************************************************/

#include <linux/module.h>
#include <linux/modversions.h> // essential, or symbols from kernel (i.e. kmalloc) and usbcore will not be recognized by depmod
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/usb.h>
#include <linux/slab.h>
#include <linux/sound.h>
#include <asm/uaccess.h>
#include <linux/soundcard.h>
#include <math.h>
#include "mmusbaudio.h"

/************************* Version Info *****************************/

#define DRIVER_AUTHOR "(c) 2001-2004 by Michael Minn"
#define DRIVER_DESC "MMUSBAUDIO: Linux USB Driver for Roland MIDI/Audio interfaces"
#define DRIVER_VERSION "v2004.10.16"

#define MMUSBAUDIO_MAX_MIDI_DEVICES 8
#define MMUSBAUDIO_MAX_AUDIO_DEVICES 8
#define MMUSBAUDIO_AUDIO_URBS 64 /* smaller numbers result in occasional pauses due to delays reading from disk? */



/******************************* USB Audio *******************************/

struct mmusbaudio_audio_device
{
	int channels;
	int resolution;
	int sample_rate;
	int device_rate;
	int device_channels;
	int device_resolution;

	int read_interface;
	int write_interface;
	int read_alternate;
	int write_alternate;
	int read_endpoint;
	int write_endpoint;
	int read_packet_size;
	int write_packet_size;

	int oss_minor_number;
	char *device_name;

	struct urb *read_urb[MMUSBAUDIO_AUDIO_URBS];
	int current_read_urb;
	int current_read_offset;
	spinlock_t read_lock;

	struct urb *write_urb[MMUSBAUDIO_AUDIO_URBS];
	int current_write_urb;
	int current_write_offset;
	spinlock_t write_lock;

	wait_queue_head_t read_wait_queue;
	wait_queue_head_t write_wait_queue;

	struct usb_device *usb_device;
} mmusbaudio_audio_devices[MMUSBAUDIO_MAX_AUDIO_DEVICES];

int mmusbaudio_find_audio_endpoints(struct mmusbaudio_audio_device *device, struct usb_device *usb_device)
{
	int scan = 0;
	int count = 0;
	unsigned char buffer[256];

	// The supported devices only have one configuration
	// Reading the configuration descriptor reads in all interface and endpoint descriptors as well

	int length = usb_control_msg(usb_device, usb_rcvctrlpipe(usb_device, 0),
		USB_REQ_GET_DESCRIPTOR, USB_DIR_IN, (USB_DT_CONFIG << 8), 0, buffer, 256, 100);

	if (length < 0)
	{
		err("Failure %d reading configuration descriptors from the device", length);
		return length;
	}

	// Determines device info based on max packet size of endpoints

	for (count = 0, scan = 0; (count < 50) && (scan < length); ++count)
	{
		int descriptor_length = buffer[scan];
		unsigned char descriptor_type = buffer[scan + 1];
		unsigned char endpoint_address = buffer[scan + 2];
		short packet_size = *((short*) &buffer[scan + 4]);

		// info("   Descriptor %d, size %d, endpoint %x", descriptor_type, packet_size, endpoint_address);

		if ((descriptor_type == 5) && (packet_size == 288))
		{
			info("UA-5 at 44.1 Khz");
			device->write_interface = 1;
			device->write_alternate = 1;
			device->write_endpoint = 1;
			device->write_packet_size = 270;
			device->read_interface = 2;
			device->read_alternate = 1;
			device->read_endpoint = 82;
			device->read_packet_size = 270;
			device->device_rate = 44100;
			device->device_channels = 2;
			device->device_resolution = 3;
			return packet_size;
		}

		else if ((descriptor_type == 5) && (packet_size == 320))
		{
			info("UA-5 at 48 Khz");
			device->write_interface = 1;
			device->write_alternate = 1;
			device->write_endpoint = 1;
			device->write_packet_size = 288;
			device->read_interface = 2;
			device->read_alternate = 1;
			device->read_endpoint = 82;
			device->read_packet_size = 288;
			device->device_rate = 48000;
			device->device_channels = 2;
			device->device_resolution = 3;
			return packet_size;
		}

		else if ((descriptor_type == 5) && (packet_size == 608) && (endpoint_address == 0x01))
		{
			info("UA-5 96 Khz playback only");
			device->write_interface = 1;
			device->write_alternate = 1;
			device->write_endpoint = 1;
			device->write_packet_size = 576;
			device->read_interface = 0;
			device->read_alternate = 0;
			device->read_endpoint = 0;
			device->read_packet_size = 0;
			device->device_rate = 96000;
			device->device_channels = 2;
			device->device_resolution = 3;
			return packet_size;
		}

		else if ((descriptor_type == 5) && (packet_size == 608) && (endpoint_address == 0x82))
		{
			info("UA-5 96 Khz record only");
			device->write_interface = 0;
			device->write_alternate = 0;
			device->write_endpoint = 0;
			device->write_packet_size = 0;
			device->read_interface = 1;
			device->read_alternate = 1;
			device->read_endpoint = 82;
			device->read_packet_size = 576;
			device->device_rate = 96000;
			device->device_channels = 2;
			device->device_resolution = 3;
			return packet_size;
		}

		else if ((descriptor_type == 5) && (packet_size == 360) && (endpoint_address == 0x01))
		{
			info("UA-100 at 44.1 Khz");
			device->write_interface = 0;
			device->write_alternate = 1;
			device->write_endpoint = 1;
			device->write_packet_size = 360;
			device->read_interface = 1;
			device->read_alternate = 1;
			device->read_endpoint = 81;
			device->read_packet_size = 184;
			device->device_rate = 44100;
			device->device_channels = 4;
			device->device_resolution = 2;
			return packet_size;
		}

		if (descriptor_length <= 0)
			scan = length;
		else
			scan += descriptor_length;
	}

	return 0;
}


static int mmusbaudio_register_sound_dsp(struct file_operations *fops);
static void mmusbaudio_unregister_sound_dsp(int unit);
struct file_operations mmusbaudio_audio_operations;

static struct mmusbaudio_audio_device *mmusbaudio_assign_audio_device(struct usb_device *usb_device, char *device_name)
{
	int scan = 0;
	int status = 0;
	int device_number = 0;
	struct mmusbaudio_audio_device *device = NULL;

	if ((status = mmusbaudio_register_sound_dsp(&mmusbaudio_audio_operations)) < 0)
	{
		err("Failure %d registering DSP device with OSS", status);
		return NULL;
	}

	device_number = status >> 4;
	if (device_number > MMUSBAUDIO_MAX_AUDIO_DEVICES)
	{
		mmusbaudio_unregister_sound_dsp(device_number);
		err("Device number %d out of range", status);
		return NULL;
	}

	device = &mmusbaudio_audio_devices[device_number];
	if (device->usb_device)
	{
		mmusbaudio_unregister_sound_dsp(device_number);
		err("Device number %d in use", status);
		return NULL;
	}

	if (mmusbaudio_find_audio_endpoints(device, usb_device) < 0)
		return NULL;
	
	if (device_number == 0)
		info("Registered %s as /dev/dsp", device_name);
	else
		info("Registered %s as /dev/dsp%d", device_name, device_number);

	device->usb_device = usb_device;
	device->device_name = device_name;
	device->oss_minor_number = status;
	device->write_lock = SPIN_LOCK_UNLOCKED;
	device->read_lock = SPIN_LOCK_UNLOCKED;

	for (scan = 0; scan < MMUSBAUDIO_AUDIO_URBS; ++scan)
		if (device->write_packet_size <= 0)
			device->write_urb[scan] = NULL;
		else
		{
			device->write_urb[scan] = usb_alloc_urb(1);
			device->write_urb[scan]->context = device;
			device->write_urb[scan]->transfer_buffer = kmalloc(device->write_packet_size, GFP_KERNEL);
			device->write_urb[scan]->transfer_buffer_length = device->write_packet_size;
			device->write_urb[scan]->actual_length = device->write_packet_size;
			device->write_urb[scan]->number_of_packets = 1;
			device->write_urb[scan]->iso_frame_desc[0].offset = 0;
			device->write_urb[scan]->iso_frame_desc[0].length = device->write_packet_size;
			device->write_urb[scan]->iso_frame_desc[0].actual_length = device->write_packet_size;
			device->write_urb[scan]->iso_frame_desc[0].status = 0;
		}


	for (scan = 0; scan < MMUSBAUDIO_AUDIO_URBS; ++scan)
		if (device->read_packet_size <= 0) 
			device->read_urb[scan] = NULL;
		else
		{
			device->read_urb[scan] = usb_alloc_urb(1);
			device->read_urb[scan]->context = device;
			device->read_urb[scan]->transfer_buffer = kmalloc(device->read_packet_size, GFP_KERNEL);
			device->read_urb[scan]->transfer_buffer_length = device->read_packet_size;
			device->read_urb[scan]->actual_length = device->read_packet_size;
			device->read_urb[scan]->number_of_packets = 1;
			device->read_urb[scan]->iso_frame_desc[0].offset = 0;
			device->read_urb[scan]->iso_frame_desc[0].length = device->read_packet_size;
			device->read_urb[scan]->iso_frame_desc[0].actual_length = device->read_packet_size;
			device->read_urb[scan]->iso_frame_desc[0].status = 0;
		}

	return device;
}

static int mmusbaudio_release_audio_device(struct mmusbaudio_audio_device *device)
{
	int scan = 0;

	// Device will be released multiple times for unused interfaces
	//if (device->device_channels <= 0)
	//	return -EINVAL;

	info("Disconnecting AUDIO device 0x%02x", device->oss_minor_number >> 4);

	device->usb_device = NULL;
	device->channels = 0;
	mmusbaudio_unregister_sound_dsp(device->oss_minor_number);

	for (scan = 0; scan < MMUSBAUDIO_AUDIO_URBS; ++scan)
	{
		if (device->read_urb[scan])
		{
			usb_unlink_urb(device->read_urb[scan]);
			kfree(device->read_urb[scan]->transfer_buffer);
			usb_free_urb(device->read_urb[scan]);
		}
		if (device->write_urb[scan])
		{
			usb_unlink_urb(device->write_urb[scan]);
			kfree(device->write_urb[scan]->transfer_buffer);
			usb_free_urb(device->write_urb[scan]);
		}
	}

	return 0;
}

static int mmusbaudio_audio_open (struct inode *inode, struct file *file)
{
	// int scan = 0;
	int retry = 10;
	int status = 0;
	int blank_urb = 0;
	int device_number = MINOR(inode->i_rdev) >> 4;
	struct mmusbaudio_audio_device *device = &mmusbaudio_audio_devices[device_number];

	info("MMUSBAUDIO Wave open device minor number %x", MINOR(inode->i_rdev));

	/****
	for (scan = 0; (scan < MMUSBAUDIO_MAX_AUDIO_DEVICES) && (!device); ++scan)
		if (MINOR(inode->i_rdev) == mmusbaudio_audio_devices[scan].oss_minor_number)
			device = &mmusbaudio_audio_devices[scan];

	if (!device)
	{
		err("Invalid MMUSBAUDIO device minor number %d", MINOR(inode->i_rdev));
		return -EINVAL;
	}
	******/
	
	// Initialize them early to prevent kernel oops if there are other problems
	init_waitqueue_head(&device->read_wait_queue);
	init_waitqueue_head(&device->write_wait_queue);

	if (!device->usb_device)
	{
		err("no device");
		return -EHOSTDOWN;
	}

	if (device->channels)
	{
		err("Open request for busy DSP device %d", device_number);
		return -EBUSY;
	}

	/* For some reason, these calls sporadically return -110 (ETIMEDOUT), so...try multiple times */

	while (retry && ((status = usb_set_interface(device->usb_device, device->read_interface, device->read_alternate)) < 0))
	{
		interruptible_sleep_on_timeout(&device->write_wait_queue, HZ / 10);
		--retry;
	}

	if (status < 0)
	{
		err("Failure %d setting read interface", status);
		return status;
	}

	while (retry && ((status = usb_set_interface(device->usb_device, device->write_interface, device->write_alternate)) < 0))
	{
		interruptible_sleep_on_timeout(&device->write_wait_queue, HZ / 10);
		--retry;
	}

	if (status < 0)
	{
		err("Failure %d setting write interface", status);
		return status;
	}

	file->private_data = (void*) MINOR(inode->i_rdev);
	device->channels = 2;
	device->sample_rate = 44100;
	device->resolution = 2;
	device->current_read_urb = 0;
	device->current_read_offset = 0;
	device->current_write_urb = 0;
	device->current_write_offset = 0;

	/* set packets initially empty so the first read immediately queues all buffers */
	for (blank_urb = 0; (blank_urb < MMUSBAUDIO_AUDIO_URBS) && device->read_urb[blank_urb] && device->read_urb[blank_urb]->transfer_buffer; 
			++blank_urb)
	{
		memset(device->read_urb[blank_urb]->transfer_buffer, 0, device->read_packet_size);
		device->read_urb[blank_urb]->iso_frame_desc[0].actual_length = 0;
	}

	MOD_INC_USE_COUNT;

	return 0;
}

static int mmusbaudio_audio_release (struct inode *inode, struct file *file)
{
	int scan;
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_audio_device *device = &mmusbaudio_audio_devices[device_number];

	info("MMUSBAUDIO Wave Close");

	if (device)
		device->channels = 0;

	for (scan = 0; device && device->usb_device && (scan < MMUSBAUDIO_AUDIO_URBS); ++scan)
	{
		if (device->read_urb && device->read_urb[scan])
			usb_unlink_urb(device->read_urb[scan]);
		if (device->write_urb && device->write_urb[scan])
			usb_unlink_urb(device->write_urb[scan]);
	}

	MOD_DEC_USE_COUNT;

	return 0;
}

static int mmusbaudio_copy_from_user(struct mmusbaudio_audio_device *device, char *current_buffer, const char *data, float *samples_transferred)
{
	int channel;
	int bytes_per_sample = device->resolution * device->channels;
	unsigned char *source = (unsigned char*) &data[(int)(*samples_transferred) * bytes_per_sample];

	if ((device->sample_rate == device->device_rate) 
		&& (device->channels == device->device_channels) 
		&& (device->resolution == device->device_resolution))
	{
		copy_from_user(current_buffer, source, bytes_per_sample);
		++(*samples_transferred);
		return bytes_per_sample;
	}

	else if (device->resolution == 3) // 24-bit
	{
		for (channel = 0; channel < device->device_channels; ++channel)
			if (device->device_resolution == 3)
				copy_from_user(&current_buffer[3 * channel], &source[3 * (channel % device->channels)], 3);

			else // 16-bit
				copy_from_user(&current_buffer[2 * channel], &source[1 + (3 * (channel % device->channels))], 2);
	}

	else if (device->resolution == 2) // 16-bit
	{
		for (channel = 0; channel < device->device_channels; ++channel)
			if (device->device_resolution == 3)
			{
				int index = 3 * channel;
				current_buffer[index] = 0;
				copy_from_user(&current_buffer[index + 1], &source[2 * (channel % device->channels)], 2);
			}

			else // 16-bit
				copy_from_user(&current_buffer[2 * channel], &source[2 * (channel % device->channels)], 2);
	}

	else // 8-bit
	{
		for (channel = 0; channel < device->device_channels; ++channel)
		{
			unsigned char char_value;
			copy_from_user(&char_value, &source[channel % device->channels], 1);
			if (device->device_resolution == 3)
			{
				int value = ((int)char_value - 0x80) << 15;
				memcpy(&current_buffer[3 * channel], &value, 3);
			}
			else // 16-bit
			{
				int value = ((int)char_value - 0x80) << 7;
				memcpy(&current_buffer[2 * channel], &value, 2);
			}
		}
	}

	(*samples_transferred) += (device->sample_rate / (float) device->device_rate);
	return bytes_per_sample;
}

static void mmusbaudio_audio_write_callback(struct urb *urb)
{
	struct mmusbaudio_audio_device *device = (struct mmusbaudio_audio_device *) urb->context;

	if (waitqueue_active(&device->write_wait_queue))
		wake_up_interruptible(&device->write_wait_queue);
}

static ssize_t mmusbaudio_audio_write (struct file *file, const char *data, size_t data_size, loff_t *offset)
{
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_audio_device *device = &mmusbaudio_audio_devices[device_number];
	int samples = data_size / device->resolution / device->channels;
	int uneven_packets = ((device->write_packet_size == 360) || (device->write_packet_size == 270));
	int bytes_per_sample = device->device_resolution * device->device_channels;
	float samples_transferred = 0;
	unsigned long flags;

	//info("write(%d)", data_size);
	
	if ((!device) || (!device->usb_device) || (device->write_packet_size <= 0))
		return -EHOSTDOWN;

	else if (data_size < (device->channels * 2))
		return -EINVAL;

	spin_lock_irqsave(&device->write_lock, flags);

	while (samples_transferred < samples)
	{
		int retry = 40;
		int current_urb_length = device->write_packet_size;
		struct urb *current_urb = device->write_urb[device->current_write_urb % MMUSBAUDIO_AUDIO_URBS];
		char *current_buffer = &((char*)current_urb->transfer_buffer)[device->current_write_offset];

		if (uneven_packets && (device->current_write_urb % 10))
			current_urb_length = device->write_packet_size - bytes_per_sample;

		/* Changing windows in KDE (page fault?) was causing this to wake up before URB */
		/* returned causing a false timeout condition.  This causes occasional playback */
		/* pauses, but at least writes don't fail and stop playback */

		while ((current_urb->status == -EINPROGRESS) && retry && device->usb_device)
		{
			interruptible_sleep_on_timeout(&device->write_wait_queue, HZ / 5);
			--retry;
			//info("Retry %d %x", retry, current_urb->status);
		}
		
		if (!device->usb_device)
		{
			err("Device disconnected during audio write wait");
			spin_unlock_irqrestore(&device->write_lock, flags);
			return -EHOSTDOWN;
		}

		if (current_urb->status == -EINPROGRESS)
		{
			err("Timed out writing %d of %d samples", (int) samples_transferred, samples);
			spin_unlock_irqrestore(&device->write_lock, flags);
			return -ETIMEDOUT;
		}

		mmusbaudio_copy_from_user(device, current_buffer, data, &samples_transferred);
		device->current_write_offset += bytes_per_sample;

		if (device->current_write_offset >= current_urb_length)
		{
			int status = 0;
			int retry = 10;

			// if ((device->current_write_urb % 100) == 0)
			// info("URB length %d %d", current_urb_length, device->current_write_offset);

			device->current_write_offset = 0;
			++device->current_write_urb;

			spin_lock_init(&current_urb->lock);
			current_urb->status = 0;
			current_urb->context = device;
			current_urb->dev = device->usb_device;
			current_urb->transfer_flags = USB_ISO_ASAP;
			current_urb->complete = mmusbaudio_audio_write_callback;
			current_urb->pipe = usb_sndisocpipe(device->usb_device, device->write_endpoint);

			current_urb->transfer_buffer_length = current_urb_length;
			current_urb->actual_length = current_urb_length;
			current_urb->number_of_packets = 1;
			current_urb->iso_frame_desc[0].offset = 0;
			current_urb->iso_frame_desc[0].length = current_urb_length;
			current_urb->iso_frame_desc[0].actual_length = current_urb_length;
			current_urb->iso_frame_desc[0].status = 0;

			/* usb_submit_urb() used to sporadically return -EAGAIN */

			while (((status = usb_submit_urb(current_urb)) == -EAGAIN) && retry)
				--retry;

			if (status != 0)
				err("failure %d submitting write_urb", status);
		}
	}

	spin_unlock_irqrestore(&device->write_lock, flags);

	return data_size;
}

static void mmusbaudio_audio_read_callback(struct urb *urb)
{
	struct mmusbaudio_audio_device *device = (struct mmusbaudio_audio_device *) urb->context;

	/* info("Read URB callback status %d", urb->status);*/
	if (waitqueue_active(&device->read_wait_queue))
		wake_up_interruptible(&device->read_wait_queue);
}

static ssize_t mmusbaudio_audio_read (struct file *file, char *data, size_t size, loff_t *offset)
{
	unsigned long flags;
	int user_data_transferred = 0;
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_audio_device *device = &mmusbaudio_audio_devices[device_number];

	if ((!device) || (!device->usb_device) || (device->read_packet_size <= 0))
		return -EHOSTDOWN;

	else if (size < 4)
		return -EINVAL;

	spin_lock_irqsave(&device->read_lock, flags);

	while (user_data_transferred < size)
	{
		int retry = 100;
		struct urb *current_urb = device->read_urb[device->current_read_urb % MMUSBAUDIO_AUDIO_URBS];

		/* As with writes, I think changing windows in KDE (page fault?) was causing */
		/* this to wake up before URB returned causing a false timeout condition.    */

		while (retry && (current_urb->status == -EINPROGRESS) && device->usb_device)
		{
			interruptible_sleep_on_timeout(&device->read_wait_queue, HZ / 5);
			/* current_urb = device->read_urb[device->current_read_urb];*/
			--retry;
		}

		if (!device->usb_device)
		{
			err("Device disconnected during audio read wait");
			spin_unlock_irqrestore(&device->read_lock, flags);
			return -EHOSTDOWN;
		}

		if (current_urb->status == -EINPROGRESS)
		{
			err("Timed out reading %d of %d bytes", user_data_transferred, size);
			spin_unlock_irqrestore(&device->read_lock, flags);
			return -ETIMEDOUT;
		}

		// Need to add sample rate conversion here

		if (device->device_channels == 4) // UA-100
		{
			copy_to_user(&data[user_data_transferred], &((char*)current_urb->transfer_buffer)[device->current_read_offset], 4);
			device->current_read_offset += 4;
			user_data_transferred += 4;
		}
		else if (device->resolution == 3) // UA-5, 24-bit
		{
			copy_to_user(&data[user_data_transferred], &((char*)current_urb->transfer_buffer)[device->current_read_offset], 6);
			device->current_read_offset += 6;
			user_data_transferred += 6;
		}
		else // UA-5, 16-bit
		{
			copy_to_user(&data[user_data_transferred], &((char*)current_urb->transfer_buffer)[device->current_read_offset + 1], 2);
			copy_to_user(&data[user_data_transferred + 2], &((char*)current_urb->transfer_buffer)[device->current_read_offset + 4], 2);
			device->current_read_offset += 6;
			user_data_transferred += 4;
		}

		/* When done reading the data from a URB - resubmit it */
		if (device->current_read_offset >= current_urb->iso_frame_desc[0].actual_length)
		{
			int status = 0;

			device->current_read_offset = 0;
			++device->current_read_urb;

			spin_lock_init(&current_urb->lock);

			current_urb->status = 0;
			current_urb->start_frame = 0;
			current_urb->context = device;
			current_urb->dev = device->usb_device;
			current_urb->transfer_flags = USB_ISO_ASAP;
			current_urb->complete = mmusbaudio_audio_read_callback;
			current_urb->pipe = usb_rcvisocpipe(device->usb_device, device->read_endpoint);

			if ((status = usb_submit_urb(current_urb)) != 0)
			{
				err("failure %d submitting read urb", status);
				spin_unlock_irqrestore(&device->read_lock, flags);
				return status;
			}
		}
	}

	spin_unlock_irqrestore(&device->read_lock, flags);
	return size;
}

static int mmusbaudio_audio_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	int data_value = 0;
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_audio_device *device = &mmusbaudio_audio_devices[device_number];

	//info("ioctl(%x)", command);
	
	if (data)
		data_value = *((int*)data);

	if (command == SNDCTL_DSP_SETFMT)
	{
		if (data_value == AFMT_U8)
			device->resolution = 1;
		else if (data_value == AFMT_S16_LE)
			device->resolution = 2;
		else if (data_value == AFMT_S24_LE)
			device->resolution = 3;
		else
		{
			err("Invalid DSP format %x", data_value);
			return -EINVAL;
		}

		info("%d bit device", device->resolution * 8);
		return 0;
	}

	else if (command == SOUND_PCM_WRITE_CHANNELS)
	{
		info("%d channels", data_value);
		device->channels = data_value;
		return 0;
	}

	else if ((command == SNDCTL_DSP_SPEED) || (command == SOUND_PCM_WRITE_RATE))
	{
		info("Bit rate %d", data_value);
		device->sample_rate = data_value;
		return 0;
	}

	// else if (command == 0x5401)
	//	return 0;

	else if (command == SOUND_PCM_READ_RATE)
		return (data = 44100);

	else if (command == SOUND_PCM_READ_CHANNELS)
	{
		info("%d channels", data_value);
		if (data > device->device_channels)
			return -EINVAL;
		else
			return 0;
	}

	else if (command == SOUND_PCM_READ_BITS)
		return (data = 16);

	else if (command == SNDCTL_DSP_GETFMTS)
	{
		*((int*)data) = AFMT_S16_LE; /* This is a mask needed by the shell play command */
		return 0;
	}

	else if (command == SNDCTL_DSP_RESET)
		return 0; /* don't do anything, but acknowledge so sox will work */

	else if (command == SNDCTL_DSP_SYNC)
		return 0; /* don't do anything, but acknowledge so sox will work */

	else if (command == SNDCTL_DSP_STEREO)
	{
		if (data_value)
		{
			info("Stereo");
			device->channels = 2;
		}
		else
		{
			info("Mono");
			device->channels = 1;
		}
		return 0; 
	}

	else if (command == SNDCTL_DSP_SETFRAGMENT)
		return 0; /* don't do anything, but acknowledge so standard audio applications will work */

	else if (command == SNDCTL_DSP_GETBLKSIZE)
	{
		*((int*)data) = device->write_packet_size * MMUSBAUDIO_AUDIO_URBS;
		return 0;
	}

	else if (command == SNDCTL_DSP_GETODELAY)
	{
		// Used by RealPlayer - just ignore and still works
		// int *delay = (int*)data;
		// info("GETODELAY %d", *delay);
		// *delay = 1000 * MMUSBAUDIO_AUDIO_WRITE_URB_SHORT / (44100 * 4);
		return 0;
	}

	else if (command == SNDCTL_DSP_GETOSPACE)
	{
		// Used by RealPlayer - not exactly sure how, but this seems to work
		int scan = device->current_write_urb;
		int end = scan + MMUSBAUDIO_AUDIO_URBS;
		
		audio_buf_info *info = (audio_buf_info *)data;
		memset(info, 0, sizeof(*info));

		while ((scan < end) && (device->write_urb[scan % MMUSBAUDIO_AUDIO_URBS]->status != -EINPROGRESS))
		{
			++scan;
			++info->fragments;
		}

		// Arbitrary limit that seems to work with mplayer
		if (info->fragments > 10)
			info->fragments = 10;

		info->fragstotal = MMUSBAUDIO_AUDIO_URBS;
		info->fragsize = device->write_packet_size;
		info->bytes = info->fragsize * info->fragments;
		// info("GETIOSPACE = %d", info->bytes);

		/**
		struct urb *current_urb = device->write_urb[device->current_write_urb % MMUSBAUDIO_AUDIO_URBS];
		audio_buf_info *info = (audio_buf_info *)data;
		
		if (current_urb->status == -EINPROGRESS)
			memset(info, 0, sizeof(*info));

		else
		{
			info->fragments = 1;
			info->fragstotal = 1;
			//info->fragsize = device->write_packet_size * MMUSBAUDIO_AUDIO_URBS;
			info->fragsize = device->write_packet_size;
			info->bytes = info->fragsize;
		}
		**/
		
		return 0;
	}

	else if (command == SNDCTL_MIDI_INFO)
	{
		struct midi_info midi_info;
		memset(&midi_info, 0, sizeof(struct midi_info));
		strcpy(midi_info.name, device->device_name);
		copy_to_user(((struct midi_info*) data), &midi_info, sizeof(struct midi_info));
		return 0;
	}

	else
	{
		info("Undefined audio IOCTL %x", command);
		return -1;
	}

}

struct file_operations mmusbaudio_audio_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	mmusbaudio_audio_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	mmusbaudio_audio_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	mmusbaudio_audio_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	mmusbaudio_audio_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	mmusbaudio_audio_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};


/************************* MIDI Event Functions *****************************/

#define MMUSBAUDIO_MIDI_NOTE_OFF 0x80
#define MMUSBAUDIO_MIDI_NOTE_ON  0x90
#define MMUSBAUDIO_MIDI_POLYPHONIC_KEY_PRESSURE 0xA0
#define MMUSBAUDIO_MIDI_CONTROL_CHANGE 0xB0
#define MMUSBAUDIO_MIDI_PROGRAM_CHANGE 0xC0
#define MMUSBAUDIO_MIDI_AFTERTOUCH 0xD0
#define MMUSBAUDIO_MIDI_PITCH_BEND 0xE0
#define MMUSBAUDIO_MIDI_SYSTEM_EXCLUSIVE 0xF0
#define MMUSBAUDIO_MIDI_TIME_CODE_QUARTER_FRAME 0xF1
#define MMUSBAUDIO_MIDI_SONG_POSITION 0xF2
#define MMUSBAUDIO_MIDI_SONG_SELECT 0xF3
#define MMUSBAUDIO_MIDI_TUNE_REQUEST 0xF6
#define MMUSBAUDIO_MIDI_END_OF_EXCLUSIVE 0xF7
#define MMUSBAUDIO_MIDI_TIMING_CLOCK 0xF8
#define MMUSBAUDIO_MIDI_START 0xFA
#define MMUSBAUDIO_MIDI_CONTINUE 0xFB
#define MMUSBAUDIO_MIDI_STOP 0xFC
#define MMUSBAUDIO_MIDI_ACTIVE_SENSING 0xFE
#define MMUSBAUDIO_MIDI_META_EVENT 0xFF

static int mmusbaudio_midi_message_length(unsigned char *message, int message_length)
{
	if (!(message[0] & 0x80))
		return message_length; /* if the message does not begin with a status byte, assume it's incomplete and needs to be sent immediately */

	else switch (message[0] & 0xF0)
	{
		case MMUSBAUDIO_MIDI_NOTE_OFF:
		case MMUSBAUDIO_MIDI_NOTE_ON:
		case MMUSBAUDIO_MIDI_POLYPHONIC_KEY_PRESSURE:
		case MMUSBAUDIO_MIDI_CONTROL_CHANGE:
		case MMUSBAUDIO_MIDI_PITCH_BEND:
			return 3;

		case MMUSBAUDIO_MIDI_PROGRAM_CHANGE:
		case MMUSBAUDIO_MIDI_AFTERTOUCH:
			return 2;

		default:
			switch (message[0])
			{
				case MMUSBAUDIO_MIDI_SONG_POSITION:
					return 3;

				case MMUSBAUDIO_MIDI_TIME_CODE_QUARTER_FRAME:
				case MMUSBAUDIO_MIDI_SONG_SELECT:
					return 2;

				case MMUSBAUDIO_MIDI_TUNE_REQUEST:
				case MMUSBAUDIO_MIDI_END_OF_EXCLUSIVE:
				case MMUSBAUDIO_MIDI_TIMING_CLOCK:
				case MMUSBAUDIO_MIDI_START:
				case MMUSBAUDIO_MIDI_CONTINUE:
				case MMUSBAUDIO_MIDI_STOP:
				case MMUSBAUDIO_MIDI_ACTIVE_SENSING:
					return 1;

				case MMUSBAUDIO_MIDI_META_EVENT:
					if (message_length <= 2)
						return 3; /* a meta event is at least three bytes long */
					else
						return message[2] + 3;

				case MMUSBAUDIO_MIDI_SYSTEM_EXCLUSIVE:
					if (message_length <= 0)
						return 2; /* Sys Ex is at least two bytes long */
					else if (message[message_length - 1] == MMUSBAUDIO_MIDI_END_OF_EXCLUSIVE)
						return message_length;
					else
						return message_length + 1;
			}
	}

	return message_length;
}


/************************* USB Debug Functions *****************************

static void mmusbaudio_urb_dump(char *message, struct urb *urb)
{
	unsigned char *char_data = (unsigned char*) urb->transfer_buffer;
	int scan = 0;

	info("%s", message);
	if (!urb)
	{
		info("NULL URB");
		return;
	}

	info(" pipe %x, transfer_flags %x", urb->pipe, urb->transfer_flags);
	printk(" status %d, length %d, actual %d", urb->status, urb->transfer_buffer_length, urb->actual_length);

	for (scan = 0; scan < urb->transfer_buffer_length; ++scan)
	{
		printk (" %02x", (int) char_data[scan]);
		if ((scan & 0xf) == 0xf)
			printk("");
  	}

	printk("");
}


static void mmusbaudio_device_dump(char *message, struct usb_device *device)
{
	info("%s", message);
	info("  Device #%d, slow %d", device->devnum, device->slow);
	info("  Alternate #%d, endpoints %d",
		device->actconfig->interface[2].act_altsetting,
		device->actconfig->interface[2].altsetting[device->actconfig->interface->act_altsetting].bNumEndpoints);
}

int mmusbaudio_scan_descriptors(struct usb_device *device)
{
	int scan = 0;
	int dump = 0;
	int count = 0;
	int length = 0;
	unsigned char buffer[256];
	char string[256];

	length = usb_control_msg(device, usb_rcvctrlpipe(device, 0),
	USB_REQ_GET_DESCRIPTOR, USB_DIR_IN, (USB_DT_CONFIG << 8), 0, buffer, 256, 100);

	info("Descriptors = %d bytes", length);

	for (count = 0; (count < 50) && (scan < length); ++count)
	{
		info("  %d) Descriptor: length %d, type %02x", scan, (int) buffer[scan], (int) buffer[scan + 1]);

		for (dump = 0, string[0] = 0; (dump < buffer[scan]) && (dump < 16); ++dump)
			sprintf(&string[strlen(string)], "%02x ", buffer[scan + dump]);

		info("  %s", string);

		if (buffer[scan] <= 0)
			scan = length;
		else
			scan += buffer[scan];
	}

	return 0;
}

**********************************************************************************/


/************************ Roland MIDI Protocol ***********************************
*
*	Although the endpoints do not indicate it, the supported Roland MIDI
*	devices use the USB MIDI class protocol.
*
*	MIDI data is sent and received in 4-byte chunks with a preceeding "burst"
*	byte that indicates the destination/source port for following 3 bytes of
*	data. All MIDI messages under status 0xf0 (SYSEX) are two or three bytes long
*	and fit in a single 4-byte chunk. Messages over status byte 0xf0 are split into
*	multiple consecutive chunks with the appropriate burst bytes specified below.
*
*	For MIDI sends, the upper 4 bits of the "burst" byte are the source/destination
*	port (described below). The lower 4 bits are either the upper 4 bits of the MIDI
*	message status byte, 4 for the beginning or middle of a SYSEX message, 6 if there
*	are two bytes in the last SYSEX chunk, and 5 if there is one byte in the last
*	chunk of a SYSEX message. For MIDI receives the protocol is basically
*	the same with the upper four bits representing the input port.
*
*		Example: 19 90 30 70 is a note on message sent to port 2.
*
*	For UA-100 sends, the "burst byte" is 0x00 for MIDI OUT 1, 0x10 for MIDI OUT 2
*	and 0x20 for the UA-100 control port which receives SYSEX messages to change
*	the internal mixer and effect settings. MIDI receives are basically the same:
*	0x00 for MIDI IN 1 and 0x00 for MIDI IN 2.
*
*	For the UM-1, there is only one input and output port, so the upper 4 bits of
*	the "burst byte" are always 0x00.
*
*	For the SC-8850, the "burst byte" indicates which of the four "parts" (i.e.
*	16-channel banks) should receive the MIDI output: 0x00 for Part A, 0x10
*	for Part B, 0x20 for Part C and 0x30 for Part D.  However, unlike the other
*	devices, for MIDI input: 0x40 for MIDI IN 1 and 0x50 for MIDI IN 2
*
*	Chunks smaller than 4 bytes (such as two-byte program change messages) are
*	padded with trailing NULLs to a 4-byte length.
*
***************************************************************************************/

struct mmusbaudio_midi_device
{
	int is_open;
	int submitted_read_urbs;
	int read_endpoint_address;
	int write_endpoint_address;
	int urb_buffer_length;
	int oss_minor_number;
	struct usb_device *usb_device;
	unsigned char burst_byte;
	struct urb *read_urb;
	struct urb *write_urb;
	wait_queue_head_t read_wait_queue;
	wait_queue_head_t write_wait_queue;
	char *device_name;
} mmusbaudio_midi_devices[MMUSBAUDIO_MAX_MIDI_DEVICES];

struct file_operations mmusbaudio_midi_operations;
static int mmusbaudio_register_sound_midi(struct file_operations *fops);
static void mmusbaudio_unregister_sound_midi(int unit);

static struct mmusbaudio_midi_device *mmusbaudio_assign_midi_device(struct usb_device *usb_device, char *device_name, 
		int in_endpoint, int out_endpoint, unsigned char burst_byte, int buffer_length)
{
	int status = 0;
	int retry = 10;
	int device_number = 0;
	struct mmusbaudio_midi_device *device = NULL;

	if ((status = mmusbaudio_register_sound_midi(&mmusbaudio_midi_operations)) < 0)
	{
		err("Failure %d registering MIDI device with OSS", status);
		return NULL;
	}

	device_number = status >> 4;
	if (device_number >= MMUSBAUDIO_MAX_MIDI_DEVICES)
	{
		err("MIDI device number %d out of range", device_number);
		mmusbaudio_unregister_sound_midi(status);
		return NULL;
	}

	device = &mmusbaudio_midi_devices[device_number];
	if (device->usb_device)
	{
		err("MIDI device number %d in use", device_number);
		mmusbaudio_unregister_sound_midi(status);
		return NULL;
	}
	
	info("Registered %s as /dev/midi%02d", device_name, device_number);

	device->oss_minor_number = status;
	device->device_name = device_name;
	device->usb_device = usb_device;
	device->read_endpoint_address = in_endpoint;
	device->write_endpoint_address = out_endpoint;
	device->burst_byte = burst_byte;
	device->urb_buffer_length = buffer_length;
	device->is_open = 0;
	device->submitted_read_urbs = 0;

	/* The buffer sent with a URB must be allocated with kmalloc */
	/* or you get an interrupt error from usb-uhci - page alignment perhaps? */
	/* For reads, the buffer must be the EXACT length specified by the */
	/* device descriptor */

	device->read_urb = usb_alloc_urb(0);
	device->read_urb->transfer_buffer = (unsigned char *)kmalloc(buffer_length, GFP_KERNEL);
	device->read_urb->context = device;
	device->read_urb->transfer_buffer_length = buffer_length;

	device->write_urb = usb_alloc_urb(0);
	device->write_urb->transfer_buffer = (unsigned char *)kmalloc(buffer_length, GFP_KERNEL);
	device->write_urb->context = device;
	device->write_urb->transfer_buffer_length = buffer_length;

	// For some reason, sporadically returns -110 ETIMEDOUT
	while (retry && ((status = usb_set_interface(usb_device, 2, 0)) < 0))
	{
		// interruptible_sleep_on_timeout(&device->write_wait_queue, HZ / 10);
		--retry;
	}

	if (!retry)
	{
		err("Failure %d setting MIDI interface", status);
		return NULL;
	}

	return device;
}

static int mmusbaudio_release_midi_device(struct mmusbaudio_midi_device *device)
{
	info("Disconnecting MIDI device 0x%02x", device->oss_minor_number >> 4);

	device->usb_device = NULL;
	device->is_open = 0;
	mmusbaudio_unregister_sound_midi(device->oss_minor_number);

	if (device->write_urb)
	{
		usb_unlink_urb(device->write_urb);
		kfree(device->write_urb->transfer_buffer);
		usb_free_urb(device->write_urb);
	}

	if (device->read_urb)
	{
		usb_unlink_urb(device->read_urb);
		kfree(device->read_urb->transfer_buffer);
		usb_free_urb(device->read_urb);
	}

	return 0;
}

static int mmusbaudio_midi_file_open (struct inode *inode, struct file *file)
{
	// int scan = 0;
	int status = 0;
	int device_number = MINOR(inode->i_rdev) >> 4;
	struct mmusbaudio_midi_device *device = &mmusbaudio_midi_devices[device_number];

	/*******
	for (scan = 0; (scan < MMUSBAUDIO_MAX_MIDI_DEVICES) && (!device); ++scan)
		if (mmusbaudio_midi_devices[scan].oss_minor_number == MINOR(inode->i_rdev))
			device = &mmusbaudio_midi_devices[scan];

	if (!device)
	{
		err("Invalid MIDI device minor number %d", MINOR(inode->i_rdev));
		return -EINVAL;
	}
	********/

	if (!device->usb_device)
	{
		err("Disconnected MIDI device minor number %d", MINOR(inode->i_rdev));
		return -EHOSTDOWN; /* device unplugged? */
	}

	if (device->is_open)
		return -EBUSY;

	info("%s opened /dev/midi%02d", device->device_name, MINOR(inode->i_rdev) >> 4);

	init_waitqueue_head(&device->read_wait_queue);
	init_waitqueue_head(&device->write_wait_queue);

	file->private_data = (void*) MINOR(inode->i_rdev);
	device->submitted_read_urbs = 0;
	device->read_urb->status = 0;
	device->write_urb->status = 0;
	++device->is_open;

	// MOD_INC_USE_COUNT;

	return status;
}

static int mmusbaudio_midi_file_release (struct inode *inode, struct file *file)
{
	int device_number = MINOR(inode->i_rdev) >> 4;
	struct mmusbaudio_midi_device *device = &mmusbaudio_midi_devices[device_number];

	info("%s closed /dev/midi%02d", device->device_name, device->oss_minor_number >> 4);

	usb_unlink_urb(device->read_urb);
	usb_unlink_urb(device->write_urb);
	--device->is_open;

	// MOD_DEC_USE_COUNT;

	return 0;
}

static void mmusbaudio_midi_write_callback(struct urb *urb)
{
	struct mmusbaudio_midi_device *device = (struct mmusbaudio_midi_device *) urb->context;

	if (waitqueue_active(&device->write_wait_queue))
		wake_up_interruptible(&device->write_wait_queue);
}

static ssize_t mmusbaudio_midi_file_write (struct file *file, const char *user_data, size_t length, loff_t *offset)
{
	int status = 0;
	int message_length = 0;
	int block_start = 0;
	unsigned char data[256];
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_midi_device *device = &mmusbaudio_midi_devices[device_number];

	if (length > 256)
		length = 256;

	copy_from_user(data, user_data, length);
	message_length = mmusbaudio_midi_message_length(data, length);

	if ((device_number >= MMUSBAUDIO_MAX_MIDI_DEVICES) || (device_number < 0))
	{
		err("Write to bad device number %d", device_number);
		return -EINVAL;
	}
	
	if ((!device) || (!device->usb_device) || (!device->is_open) || (!device->write_urb) || (!device->write_urb->transfer_buffer))
		return -EHOSTDOWN; /* device unplugged? */

	else if ((data[0] < 0x80) || (data[0] > 0xf4))
		return -EINVAL; /* ignore MIDI messages without status byte or status greater than SYSEX or Roland part select F4 */

	else if (length < message_length)
		return -EINVAL; /* Ignore incomplete MIDI messages */

	// info("MIDI write %02x to device %02x", (int) data, device_number);

	/* The UA-100 limits bulk transfers to 32 bytes per URB (i.e. 24 bytes of data + the burst bytes) */
	/* Therefore, long SYSEX messages (such as the 40 byte effect parameter messages) must be broken */
	/* into multiple successive URBs */
	for (block_start = 0; block_start < length; block_start += (device->urb_buffer_length * 3 / 4))
	{
		int chunk_start = 0;
		int block_end = block_start + (device->urb_buffer_length * 3 / 4);

		if (block_end > length)
			block_end = length;

		if (device->write_urb->status == -EINPROGRESS)
			interruptible_sleep_on_timeout(&device->write_wait_queue, HZ);

		if (device->write_urb->status == -EINPROGRESS)
			return -ETIMEDOUT;

		device->write_urb->transfer_buffer_length = 0;
		for (chunk_start = block_start; chunk_start < block_end; chunk_start += 3)
		{
			int chunk_length = 3;
			if (chunk_length > (length - chunk_start))
				chunk_length = length - chunk_start;

			if (data[0] != 0xf0)
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 
					((data[chunk_start] >> 4) & 0x0f) | device->burst_byte;
			else if ((chunk_start + 3) < length)
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 
					0x04 | device->burst_byte;
			else if ((length - chunk_start) == 1)
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 
					0x05 | device->burst_byte;
			else
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 
					0x06 | device->burst_byte;

			/* chunk length must be multiple of 4? */
			memset(&((unsigned char*)device->write_urb->transfer_buffer)[device->write_urb->transfer_buffer_length + 1], 0, 3); 
			memcpy(&((unsigned char*)device->write_urb->transfer_buffer)[device->write_urb->transfer_buffer_length + 1], 
					&data[chunk_start], chunk_length);
			device->write_urb->transfer_buffer_length += 4;
		}

		spin_lock_init(&device->write_urb->lock);
		device->write_urb->dev = device->usb_device;
		device->write_urb->pipe = usb_sndbulkpipe(device->usb_device, device->write_endpoint_address);
		device->write_urb->complete = (usb_complete_t)mmusbaudio_midi_write_callback;
		device->write_urb->context = device;

		/*mmusbaudio_urb_dump("MIDI write", device->write_urb);*/

		if ((status = usb_submit_urb(device->write_urb)) != 0)
		{
			err("failure %d submitting write_urb (%d bytes)", status, device->write_urb->transfer_buffer_length);
			return status;
		}
		else
			status = length;
	}

	return status;
}

static void mmusbaudio_midi_read_callback(struct urb *urb)
{
	struct mmusbaudio_midi_device *device = (struct mmusbaudio_midi_device *) urb->context;

	if (waitqueue_active(&device->read_wait_queue))
		wake_up_interruptible(&device->read_wait_queue);
}

static ssize_t mmusbaudio_midi_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int scan = 0;
	int status = 0;
	int actual_length = 0;
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_midi_device *device = &mmusbaudio_midi_devices[device_number];

	if ((!device) || (!device->usb_device) || (!device->is_open) || (!device->read_urb) || (!device->read_urb->transfer_buffer))
		return -EHOSTDOWN; /* device unplugged? */

	if (device->submitted_read_urbs <= 0)
	{
		spin_lock_init(&device->read_urb->lock);
		device->read_urb->dev = device->usb_device;
		device->read_urb->pipe = usb_rcvbulkpipe(device->usb_device, device->read_endpoint_address);
		((unsigned char*)device->read_urb->transfer_buffer)[0] = 0;
		device->read_urb->complete = (usb_complete_t)mmusbaudio_midi_read_callback;
		device->read_urb->context = device;
		device->submitted_read_urbs = 1;

		if ((status = usb_submit_urb(device->read_urb)) < 0)
		{
			err("failure %d submitting read_urb to endpoint %02x", status, device->read_endpoint_address);
			return status;
		}

		//interruptible_sleep_on(&device->read_wait_queue);
		interruptible_sleep_on_timeout(&device->read_wait_queue, 1);

		/* mmusbaudio_urb_dump("Read URB -----------", device->read_urb); */
	}


	if ((!device) || (!device->usb_device) || (!device->is_open))
		return -EHOSTDOWN; /* device unplugged? */

	// MMUSBAUDIO only implements nonblocking MIDI read
	// if the device does not return the URB by the timeout given above, return
	else if (device->read_urb->status == -EINPROGRESS)
		return -EAGAIN;

	// UA-100 regularly returns empty buffers
	else if (((unsigned char*)device->read_urb->transfer_buffer)[0] == 0)
		actual_length = -EAGAIN;

	/*
	else if ((((unsigned char*)device->read_urb->transfer_buffer)[0] & 0xf0) != device->burst_byte)
		length = 0; data from a different port - doesn't work with SC-8850
	*/

	else for (scan = 0; (scan < device->read_urb->actual_length) && (((scan / 4) * 3) < length) 
			&& (((unsigned char*)device->read_urb->transfer_buffer) [scan] > 0); scan += 4)
		switch (((unsigned char*)device->read_urb->transfer_buffer) [scan] & 0x0f)
		{
			case 0x05: /* 1 byte end of sysex */
			case 0x0F: /* single byte ? */
				copy_to_user(&data[(scan / 4) * 3], &((unsigned char*)device->read_urb->transfer_buffer) [scan + 1], 1);
				actual_length += 1;
				break;

			case 0x02: /* 2 byte system common message */
			case 0x06: /* 2 byte end of sysex */
			case 0x0C: /* Program Change*/
			case 0x0D: /* Aftertouch */
				copy_to_user(&data[(scan / 4) * 3], &((unsigned char*)device->read_urb->transfer_buffer) [scan + 1], 2);
				actual_length += 2;
				break;

			default: /* everything else is 3-byte packet */
				copy_to_user(&data[(scan / 4) * 3], &((unsigned char*)device->read_urb->transfer_buffer) [scan + 1], 3);
				actual_length += 3;
				break;
		}

	device->submitted_read_urbs = 0;

	return actual_length;
}

static int mmusbaudio_midi_file_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	int device_number = ((int) file->private_data) >> 4;
	struct mmusbaudio_midi_device *device = &mmusbaudio_midi_devices[device_number];
	struct midi_info midi_info;

	if (command == SNDCTL_MIDI_INFO)
	{
		memset(&midi_info, 0, sizeof(struct midi_info));
		strcpy(midi_info.name, device->device_name);
		copy_to_user(((struct midi_info*) data), &midi_info, sizeof(struct midi_info));
		return 0;
	}

	else
	{
		info("Undefined MIDI IOCTL %x", command);
		return -1;
	}
}

struct file_operations mmusbaudio_midi_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	mmusbaudio_midi_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	mmusbaudio_midi_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	mmusbaudio_midi_file_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	mmusbaudio_midi_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	mmusbaudio_midi_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};

/*********************** OSS and Standalone Compatibility Functions *************************

  MMUSBAUDIO can operate within the legacy Linux Open Sound System (OSS) or as a standalone
  module that uses the same devices (major number 14) as OSS. The following functions
  are conditionally compiled wrapper functions that allow the rest of the code to be
  independent of which type of module this is compiled as.

  Compilation for standalone is made if MMUSBAUDIO_DEVICE_MAJOR_NUMBER is defined in the
  Makefile. Otherwise, compilation assumes OSS.
  
  All OSS operations are multiplexed through the soundcore module and a set of devices
  with a single major number. Therefore, for compatibility, in standalone mode there
  is a separate set of file_operations functions that then call either the audio or
  midi file_operations functions. The four upper bits of the minor number are the 
  associated device number and the four lower bits of the minor number indicate the
  type of device (2 for MIDI, 3 for DSP).

********************************************************************************************/

#ifndef MMUSBAUDIO_DEVICE_MAJOR_NUMBER

static int mmusbaudio_register_sound_midi(struct file_operations *fops)
{
	return register_sound_midi(fops, -1);
}

static int mmusbaudio_register_sound_dsp(struct file_operations *fops)
{
	return register_sound_dsp(fops, -1);
}

static void mmusbaudio_unregister_sound_midi(int unit)
{
	unregister_sound_midi(unit);
}

static void mmusbaudio_unregister_sound_dsp(int unit)
{
	unregister_sound_dsp(unit);
}

static int mmusbaudio_register_chrdev()
{
	return 0;
}

static void mmusbaudio_unregister_chrdev()
{	
}

#else

static int mmusbaudio_register_sound_midi(struct file_operations *fops)
{
	int device_number = 0;
	
	while ((device_number < MMUSBAUDIO_MAX_MIDI_DEVICES) && (mmusbaudio_midi_devices[device_number].usb_device != NULL))
		++device_number;

	if (device_number >= MMUSBAUDIO_MAX_MIDI_DEVICES)
	{
		err("MMUSBAUDIO maximum %d MIDI devices are already connected", MMUSBAUDIO_MAX_MIDI_DEVICES);
		return -1;
	}

	return (device_number << 4) + 2;
}

static int mmusbaudio_register_sound_dsp(struct file_operations *fops)
{
	int device_number = 0;
	
	while ((device_number < MMUSBAUDIO_MAX_AUDIO_DEVICES) && (mmusbaudio_audio_devices[device_number].usb_device != NULL))
		++device_number;

	if (device_number >= MMUSBAUDIO_MAX_AUDIO_DEVICES)
	{
		err("MMUSBAUDIO maximum %d audio devices are already connected", MMUSBAUDIO_MAX_AUDIO_DEVICES);
		return -1;
	}

	return (device_number << 4) + 3;
}

static void mmusbaudio_unregister_sound_midi(int unit)
{
}

static void mmusbaudio_unregister_sound_dsp(int unit)
{
}

static int mmusbaudio_open (struct inode *inode, struct file *file)
{
	int device_type = MINOR(inode->i_rdev) & 0xf;
	if (device_type == 2)
		return mmusbaudio_midi_file_open(inode, file);
	else if (device_type == 3)
		return mmusbaudio_audio_open(inode, file);
	else
		return -ENODEV;
}

static int mmusbaudio_release (struct inode *inode, struct file *file)
{
	int device_type = MINOR(inode->i_rdev) & 0xf;
	if (device_type == 2)
		return mmusbaudio_midi_file_release(inode, file);
	else if (device_type == 3)
		return mmusbaudio_audio_release(inode, file);
	else
		return -ENODEV;
}

static ssize_t mmusbaudio_write (struct file *file, const char *user_data, size_t length, loff_t *offset)
{
	int device_type = ((int)file->private_data) & 0xf;
	if (device_type == 2)
		return mmusbaudio_midi_file_write(file, user_data, length, offset);
	else if (device_type == 3)
		return mmusbaudio_audio_write(file, user_data, length, offset);
	else
		return -ENODEV;	
}

static ssize_t mmusbaudio_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int device_type = ((int)file->private_data) & 0xf;
	if (device_type == 2)
		return mmusbaudio_midi_file_read(file, data, length, offset);
	else if (device_type == 3)
		return mmusbaudio_audio_read(file, data, length, offset);
	else
		return -ENODEV;	
}

static int mmusbaudio_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	int device_type = MINOR(inode->i_rdev) & 0xf;
	if (device_type == 2)
		return mmusbaudio_midi_file_ioctl(inode, file, command, data);
	else if (device_type == 3)
		return mmusbaudio_audio_ioctl(inode, file, command, data);
	else
		return -ENODEV;
}

struct file_operations mmusbaudio_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	mmusbaudio_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	mmusbaudio_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	mmusbaudio_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	mmusbaudio_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	mmusbaudio_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};
static int mmusbaudio_register_chrdev()
{	
	info("MMUSBAUDIO compiled for stand-alone operation (no OSS soundcore): device major number %d", MMUSBAUDIO_DEVICE_MAJOR_NUMBER);
	return register_chrdev(MMUSBAUDIO_DEVICE_MAJOR_NUMBER, "mmusbaudio", &mmusbaudio_operations);
}

static void mmusbaudio_unregister_chrdev()
{	
	unregister_chrdev(MMUSBAUDIO_DEVICE_MAJOR_NUMBER, "mmusbaudio");
}

#endif

/************************* USB Driver Registration *****************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static void *mmusbaudio_probe(struct usb_device *device, unsigned interface, const struct usb_device_id *id)
#else
static void *mmusbaudio_probe(struct usb_device *device, unsigned interface)
#endif
{
	if ((device->descriptor.idVendor == 0x582) && 
			(device->descriptor.idProduct == 0) && (interface == 0))
		return mmusbaudio_assign_audio_device(device, "UA-100 Audio");

	else if ((device->descriptor.idVendor == 0x582) && 
			(device->descriptor.idProduct == 0x10) && (interface == 1))
		return mmusbaudio_assign_audio_device(device, "UA-5");

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 0x10))
	{
		// return &mmusbaudio_audio_devices[0]; // kludge - what if more than one device?
		// The UA-5 input interface is added above; Other interfaces ignored but acknowledged to avoid error msgs
		return 0;
	}

	else if ((device->descriptor.idVendor == 0x582) && 
			(device->descriptor.idProduct == 0) && (interface == 1))
		return 0; // The UA-100 input interface - ignore because it's added above

	else if ((device->descriptor.idVendor == 0x582) && 
			(device->descriptor.idProduct == 0) && (interface == 2))
	{
		mmusbaudio_assign_midi_device(device, "UA-100 A", 0x82, 0x02, 0x00, 32);
		mmusbaudio_assign_midi_device(device, "UA-100 B", 0x82, 0x02, 0x10, 32);
		return mmusbaudio_assign_midi_device(device, "UA-100 CTRL", 0x82, 0x02, 0x20, 32);
	}

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 3) && (interface == 2))
	{
		mmusbaudio_assign_midi_device(device, "SC-8850 A", 0x81, 0x01, 0x00, 64);
		mmusbaudio_assign_midi_device(device, "SC-8850 B", 0x81, 0x01, 0x10, 64);
		mmusbaudio_assign_midi_device(device, "SC-8850 C", 0x81, 0x01, 0x20, 64);
		return mmusbaudio_assign_midi_device(device, "SC-8850 D", 0x81, 0x01, 0x30, 64);
	}

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 9) && (interface == 2))
		return mmusbaudio_assign_midi_device(device, "UM-1", 0x81, 0x01, 0x00, 64);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 0))
		info("Unsupported UA-100 interface %d", interface);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 3))
		info("Unsupported SC-8850 interface %d", interface);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 9))
		info("Unsupported UM-1 interface %d", interface);

	return NULL;
}

static void mmusbaudio_disconnect(struct usb_device *device, void *driver_context)
{
	struct mmusbaudio_midi_device *midi_device = mmusbaudio_midi_devices;
	struct mmusbaudio_audio_device *audio_device = mmusbaudio_audio_devices;
	
	while (audio_device < &mmusbaudio_audio_devices[MMUSBAUDIO_MAX_AUDIO_DEVICES])
	{
		if (audio_device->usb_device == device)
			mmusbaudio_release_audio_device(audio_device);
		++audio_device;
	}

	while (midi_device < &mmusbaudio_midi_devices[MMUSBAUDIO_MAX_AUDIO_DEVICES])
	{
		if (midi_device->usb_device == device)
			mmusbaudio_release_midi_device(midi_device);
		++midi_device;
	}
}

static struct usb_driver mmusbaudio_driver_registration =
{
	name: "mmusbaudio",
	probe: mmusbaudio_probe,
	disconnect: mmusbaudio_disconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	id_table:	NULL,
	driver_list: {NULL, NULL},
	fops: NULL
#endif
};


/************************* Module Entry and Exit *****************************/

static int __init mmusbaudio_init(void)
{
	int status = 0;

	info(DRIVER_DESC " " DRIVER_AUTHOR " " DRIVER_VERSION);

	memset(&mmusbaudio_midi_devices, 0, sizeof(mmusbaudio_midi_devices));
	memset(&mmusbaudio_audio_devices, 0, sizeof(mmusbaudio_midi_devices));
	
	if ((status = mmusbaudio_register_chrdev()) < 0)
	{
		err("MMUSBAUDIO: register_chrdev() failed code %d", status);
		return -1;
	}
	
	if ((status = usb_register(&mmusbaudio_driver_registration)) < 0)
		err("Driver failed usb_register with code %d", status);

	return status;
}

static void __exit mmusbaudio_cleanup(void)
{
	mmusbaudio_unregister_chrdev();
	
	usb_deregister(&mmusbaudio_driver_registration);

	info("Module deregistered and unloaded");
}

module_init(mmusbaudio_init);
module_exit(mmusbaudio_cleanup);

/****************** Modprobe/Hotplugging Identification Info ******************/

static struct usb_device_id mmusbaudio_ids [] =
{
	{ USB_DEVICE_VER(0x582, 0x0000, 0x0, 0xffff) }, /* UA-100  */
	{ USB_DEVICE_VER(0x582, 0x0003, 0x0, 0xffff) }, /* SC-8850 */
	{ USB_DEVICE_VER(0x582, 0x0009, 0x0, 0xffff) }, /* UM-1    */
	{ USB_DEVICE_VER(0x582, 0x0010, 0x0, 0xffff) }, /* UA-5    */
	{ } /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, mmusbaudio_ids);

MODULE_AUTHOR(DRIVER_AUTHOR);

MODULE_DESCRIPTION(DRIVER_DESC);

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

